"""
pyEQL solute properties test suite
============================================

This file contains tests of the Solution class methods that retrieve or
calculate properties of individual solutes. Methods currently included
in the testing are:

- get_transport_number()
- get_molar_conductivity()
- get_mobility()


"""

import numpy as np

from pyEQL import Solution, ureg

# relative tolerance between experimental and computed properties for this test file
RTOL = 0.05


class Test_transport_number:
    """
    test get_transport_number
    ------------------------------------------------
    """

    def test_transport_number(self):
        # create a test solution
        s1 = Solution([["K+", "0.1 mol/L"], ["Cl-", "0.1 mol/L"], ["FeO", "0.2 mol/L"]])

        # The transport number of any ion should be between 0 and 1
        assert s1.get_transport_number("K+") >= 0
        assert s1.get_transport_number("K+") <= 1

        # The transport number of water should be 0
        assert s1.get_transport_number("H2O") == 0

        # The transport number of any uncharged species should be 0
        assert s1.get_transport_number("FeO") == 0

        # The transport numbers should add up to 1
        total_t = 0
        for item in s1.components:
            total_t += s1.get_transport_number(item)

        assert round(abs(total_t.magnitude - 1), 7) == 0


class Test_molar_conductivity:
    """
    test get_molar_conductivity

    Published molar conductivity values at 25 degC and infinite dilution
    are found in CRC Handbook of Chemistry and Physics, 92nd ed., 2011.

    This reference is also the source of diffusion coefficient data
    in the default database, so the values molar conductivity (which
    is calculated from the diffusion coefficient) should match.
    ----------------------------------------------------------------
    """

    def test_molar_conductivity_potassium(self):
        # K+ - 73.48 x 10 ** -4 m ** 2 S / mol
        s1 = Solution([["K+", "0.001 mol/L"], ["Cl-", "0.001 mol/L"]], temperature="25 degC")
        result = s1.get_molar_conductivity("K+").to("m**2*S/mol").magnitude
        expected = ureg.Quantity("73.48e-4 m**2 * S / mol").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    def test_molar_conductivity_sodium(self):
        # Na+ - 50.08 x 10 ** -4 m ** 2 S / mol
        s1 = Solution([["Na+", "0.001 mol/L"], ["Cl-", "0.001 mol/L"]], temperature="25 degC")
        result = s1.get_molar_conductivity("Na+").to("m**2*S/mol").magnitude
        expected = ureg.Quantity("50.08e-4 m**2 * S / mol").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    def test_molar_conductivity_magnesium(self):
        # Mg+2 - 106 x 10 ** -4 m ** 2 S / mol
        s1 = Solution([["Mg+2", "0.001 mol/L"], ["Cl-", "0.002 mol/L"]], temperature="25 degC")
        result = s1.get_molar_conductivity("Mg+2").to("m**2*S/mol").magnitude
        expected = ureg.Quantity("106e-4 m**2 * S / mol").magnitude

        assert np.isclose(result, expected, atol=0.005)

    def test_molar_conductivity_chloride(self):
        # Cl- - 76.31 x 10 ** -4 m ** 2 S / mol
        s1 = Solution([["Na+", "0.001 mol/L"], ["Cl-", "0.001 mol/L"]], temperature="25 degC")
        result = s1.get_molar_conductivity("Cl-").to("m**2*S/mol").magnitude
        expected = ureg.Quantity("76.31e-4 m**2 * S / mol").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    def test_molar_conductivity_fluoride(self):
        # F- - 55.4 x 10 ** -4 m ** 2 S / mol
        s1 = Solution([["Na+", "0.001 mol/L"], ["F-", "0.001 mol/L"]], temperature="25 degC")
        result = s1.get_molar_conductivity("F-").to("m**2*S/mol").magnitude
        expected = ureg.Quantity("55.4e-4 m**2 * S / mol").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    def test_molar_conductivity_sulfate(self):
        # SO4-2 - 160 x 10 ** -4 m ** 2 S / mol
        s1 = Solution([["Na+", "0.002 mol/L"], ["SO4-2", "0.001 mol/L"]], temperature="25 degC")
        result = s1.get_molar_conductivity("SO4-2").to("m**2*S/mol").magnitude
        expected = ureg.Quantity("160.0e-4 m**2 * S / mol").magnitude

        assert np.isclose(result, expected, atol=0.002)

    def test_molar_conductivity_hydroxide(self):
        # OH- - 198 x 10 ** -4 m ** 2 S / mol
        s1 = Solution(temperature="25 degC")
        result = s1.get_molar_conductivity("OH-").to("m**2*S/mol").magnitude
        expected = ureg.Quantity("198e-4 m**2 * S / mol").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    def test_molar_conductivity_hydrogen(self):
        # H+ - 349.65 x 10 ** -4 m ** 2 S / mol
        s1 = Solution(temperature="25 degC")
        result = s1.get_molar_conductivity("H+").to("m**2*S/mol").magnitude
        expected = ureg.Quantity("349.65e-4 m**2 * S / mol").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    # molar conductivity of a neutral solute should be zero
    def test_molar_conductivity_neutral(self):
        s1 = Solution([["FeCl3", "0.001 mol/L"]], temperature="25 degC")
        result = s1.get_molar_conductivity("FeCl3").to("m**2*S/mol").magnitude
        expected = ureg.Quantity(0, "m**2 * S / mol").magnitude

        assert round(abs(result - expected), 5) == 0

    # molar conductivity of water should be zero
    def test_molar_conductivity_water(self):
        s1 = Solution(temperature="25 degC")
        result = s1.get_molar_conductivity("H2O").to("m**2*S/mol").magnitude
        expected = ureg.Quantity(0, "m**2 * S / mol").magnitude

        assert round(abs(result - expected), 5) == 0


class Test_mobility:
    """
    test get_mobility

    Since the mobility is simply the molar conductivity divided by
    Faraday's constant and the ion charge, and we have already
    tested get_molar_conductivity, here we only do a couple of tests
    to make sure get_mobility returns a value that is correctly
    scaled related to get_molar_conductivity
    ----------------------------------------------------------------
    """

    # relative error tolerance for assertWithinExperimentalError

    def test_mobility_potassium(self):
        s1 = Solution([["K+", "0.001 mol/L"], ["Cl-", "0.001 mol/L"]], temperature="25 degC")
        molar_conductivity = s1.get_molar_conductivity("K+").to("m**2*S/mol")
        expected = s1.get_mobility("K+").to("m**2/s/V").magnitude
        charge = s1.get_property("K+", "charge")
        # calculate the mobility from get_molar_conductivity, then compare with get_mobility
        result = (molar_conductivity / (ureg.N_A * ureg.e * abs(charge))).to("m**2/s/V").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    def test_mobility_chloride(self):
        s1 = Solution([["K+", "0.001 mol/L"], ["Cl-", "0.001 mol/L"]], temperature="25 degC")
        molar_conductivity = s1.get_molar_conductivity("Cl-").to("m**2*S/mol")
        expected = s1.get_mobility("Cl-").to("m**2/s/V").magnitude
        charge = s1.get_property("Cl-", "charge")
        # calculate the mobility from get_molar_conductivity, then compare with get_mobility
        result = (molar_conductivity / (ureg.N_A * ureg.e * abs(charge))).to("m**2/s/V").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    def test_mobility_magnesium(self):
        s1 = Solution([["Mg+2", "0.001 mol/L"], ["Cl-", "0.002 mol/L"]], temperature="25 degC")
        molar_conductivity = s1.get_molar_conductivity("Mg+2").to("m**2*S/mol")
        expected = s1.get_mobility("Mg+2").to("m**2/s/V").magnitude
        charge = s1.get_property("Mg+2", "charge")
        # calculate the mobility from get_molar_conductivity, then compare with get_mobility
        result = (molar_conductivity / (ureg.N_A * ureg.e * abs(charge))).to("m**2/s/V").magnitude

        assert np.isclose(result, expected, rtol=RTOL)

    def test_mobility_sulfate(self):
        s1 = Solution([["K+", "0.002 mol/L"], ["SO4-2", "0.001 mol/L"]], temperature="25 degC")
        molar_conductivity = s1.get_molar_conductivity("SO4-2").to("m**2*S/mol")
        expected = s1.get_mobility("SO4-2").to("m**2/s/V").magnitude
        charge = s1.get_property("SO4-2", "charge")
        # calculate the mobility from get_molar_conductivity, then compare with get_mobility
        result = (molar_conductivity / (ureg.N_A * ureg.e * abs(charge))).to("m**2/s/V").magnitude

        assert np.isclose(result, expected, rtol=RTOL)



"""
Tests of pyEQL.functions module

"""
import platform

import numpy as np
import pytest

from pyEQL import Solution
from pyEQL.functions import entropy_mix, gibbs_mix


@pytest.fixture()
def s1():
    return Solution(volume="2 L")


@pytest.fixture()
def s2():
    return Solution({"Na+": "1 mol/L", "Cl-": "1 mol/L"}, volume="10 L")


@pytest.fixture()
def s1_p():
    return Solution(volume="2 L", engine="phreeqc")


@pytest.fixture()
def s2_p():
    return Solution({"Na+": "1 mol/L", "Cl-": "1 mol/L"}, volume="10 L", engine="phreeqc")


@pytest.fixture()
def s1_i():
    return Solution(volume="2 L", engine="ideal")


@pytest.fixture()
def s2_i():
    return Solution({"Na+": "1 mol/L", "Cl-": "1 mol/L"}, volume="10 L", engine="ideal")


@pytest.mark.skipif(platform.machine() == "arm64" and platform.system() == "Darwin", reason="arm64 not supported")
def test_mixing_functions(s1, s2, s1_p, s2_p, s1_i, s2_i):
    # mixing energy and entropy of any solution with itself should be zero
    assert np.isclose(gibbs_mix(s1, s1).magnitude, 0)
    assert np.isclose(entropy_mix(s2, s2).magnitude, 0)
    assert np.isclose(gibbs_mix(s1_p, s1_p).magnitude, 0)
    assert np.isclose(entropy_mix(s2_p, s2_p).magnitude, 0)
    assert np.isclose(gibbs_mix(s1_i, s1_i).magnitude, 0, atol=1e-6)
    assert np.isclose(entropy_mix(s2_i, s2_i).magnitude, 0)

    # TODO - I have not tested how equilibrate() affects the results
    for dil, conc in zip([s1, s1_p, s1_i], [s2, s2_p, s2_i]):
        # for mixing 1 and 2, we should have
        # H20: 55.5 * 2 mol + 55.5 * 10 mol, x1 = 0.9999 x2 = 0.9645, mixture = 0.9703 = approximately -9043 J
        s_theoretical = (
            8.314
            * 298.15
            * (
                (dil + conc).get_amount("H2O", "mol").magnitude
                * np.log((dil + conc).get_amount("H2O", "fraction").magnitude)
                + (dil + conc).get_amount("Na+", "mol").magnitude
                * np.log((dil + conc).get_amount("Na+", "fraction").magnitude)
                + (dil + conc).get_amount("Cl-", "mol").magnitude
                * np.log((dil + conc).get_amount("Cl-", "fraction").magnitude)
                - dil.get_amount("H2O", "mol").magnitude * np.log(dil.get_amount("H2O", "fraction").magnitude)
                - conc.get_amount("H2O", "mol").magnitude * np.log(conc.get_amount("H2O", "fraction").magnitude)
                - conc.get_amount("Na+", "mol").magnitude * np.log(conc.get_amount("Na+", "fraction").magnitude)
                - conc.get_amount("Cl-", "mol").magnitude * np.log(conc.get_amount("Cl-", "fraction").magnitude)
            )
        )
        assert np.isclose(entropy_mix(dil, conc).magnitude, s_theoretical, rtol=0.005)
        g_theoretical = (
            8.314
            * 298.15
            * (
                (dil + conc).get_amount("H2O", "mol").magnitude * np.log((dil + conc).get_activity("H2O").magnitude)
                + (dil + conc).get_amount("Na+", "mol").magnitude * np.log((dil + conc).get_activity("Na+").magnitude)
                + (dil + conc).get_amount("Cl-", "mol").magnitude * np.log((dil + conc).get_activity("Cl-").magnitude)
                - dil.get_amount("H2O", "mol").magnitude * np.log(dil.get_activity("H2O").magnitude)
                - conc.get_amount("H2O", "mol").magnitude * np.log(conc.get_activity("H2O").magnitude)
                - conc.get_amount("Na+", "mol").magnitude * np.log(conc.get_activity("Na+").magnitude)
                - conc.get_amount("Cl-", "mol").magnitude * np.log(conc.get_activity("Cl-").magnitude)
            )
        )
        assert np.isclose(gibbs_mix(dil, conc).magnitude, g_theoretical, rtol=0.005)



"""
pyEQL Solute class.

This file contains functions and methods for managing properties of
individual solutes. The Solute class contains methods for accessing
ONLY those properties that DO NOT depend on solution composition.
Solute properties such as activity coefficient or concentration
that do depend on compsition are accessed via Solution class methods.

:copyright: 2013-2024 by Ryan S. Kingsbury
:license: LGPL, see LICENSE for more details.

"""

from __future__ import annotations

import logging
import warnings
from dataclasses import asdict, dataclass, field
from typing import Literal

import numpy as np
from pymatgen.core.ion import Ion

from pyEQL.utils import standardize_formula

logger = logging.getLogger(__name__)


@dataclass
class Datum:
    """Document containing data for a single computed or experimental property."""

    value: str
    reference: str | None = None
    data_type: Literal["computed", "experimental", "fitted", "unknown"] = "unknown"

    @property
    def magnitude(self):
        """Return the numerical value of a Datum."""
        return float(self.value.split(" ")[0])

    @property
    def unit(self):
        """Return the unit of a Datum."""
        return self.value.split(" ")[-1]

    @property
    def uncertainty(self):
        """Return the uncertainty of a Datum."""
        if len(self.value.split(" ")) > 3:
            return float(self.value.split(" ")[2])
        return np.nan

    def as_dict(self):
        """Return a dictionary representation of the Datum."""
        return dict(asdict(self).items())


@dataclass
class Solute:
    """
    represent each chemical species as an object containing its formal charge,
    transport numbers, concentration, activity, etc.

    Args:
        formula: Chemical formula for the solute. Charged species must contain a + or - and (for polyvalent solutes)
            a number representing the net charge (e.g. 'SO4-2').
    """

    formula: str
    charge: int
    molecular_weight: str
    elements: list
    chemsys: str
    pmg_ion: Ion
    formula_html: str
    formula_latex: str
    formula_hill: str
    formula_pretty: str
    oxi_state_guesses: dict[str, float]
    n_atoms: int
    n_elements: int
    size: dict = field(
        default_factory=lambda: {
            "radius_ionic": None,
            "radius_hydrated": None,
            "radius_vdw": None,
            "molar_volume": None,
        }
    )
    thermo: dict = field(default_factory=lambda: {"ΔG_hydration": None, "ΔG_formation": None})
    transport: dict = field(default_factory=lambda: {"diffusion_coefficient": None})
    model_parameters: dict = field(
        default_factory=lambda: {
            "activity_pitzer": {"Beta0": None, "Beta1": None, "Beta2": None, "Cphi": None, "Max_C": None},
            "molar_volume_pitzer": {
                "Beta0": None,
                "Beta1": None,
                "Beta2": None,
                "Cphi": None,
                "V_o": None,
                "Max_C": None,
            },
            "viscosity_jones_dole": {"B": None},
            "diffusion_temp_smolyakov": {"a1": None, "a2": None, "d": None},
        }
    )

    @classmethod
    def from_formula(cls, formula: str):
        """
        Create an Ion document from a chemical formula. The formula is passed to
        pymatgen.core.Ion.from_formula() and used to populate the basic chemical
        informatics fields (e.g., formula, charge, molecular weight, elements, etc.)
        of the IonDoc.
        """
        pmg_ion = Ion.from_formula(formula)
        f, factor = pmg_ion.get_reduced_formula_and_factor()
        rform = standardize_formula(formula)
        charge = int(pmg_ion.charge)
        els = [str(el) for el in pmg_ion.elements]
        mw = f"{float(pmg_ion.weight / factor)} g/mol"  # weight is a FloatWithUnit
        chemsys = pmg_ion.chemical_system
        # store only the most likely oxi_state guesses
        try:
            oxi_states = pmg_ion.oxi_state_guesses(all_oxi_states=True)[0]
        except (IndexError, ValueError):
            warnings.warn(f"Guessing oxi states failed for {formula}")
            oxi_states = {}

        return cls(
            rform,
            charge=charge,
            molecular_weight=mw,
            elements=els,
            chemsys=chemsys,
            pmg_ion=pmg_ion,
            formula_html=pmg_ion.to_html_string(),
            formula_latex=pmg_ion.to_latex_string(),
            formula_hill=pmg_ion.hill_formula,
            formula_pretty=pmg_ion.to_pretty_string(),
            oxi_state_guesses=oxi_states,
            n_atoms=int(pmg_ion.num_atoms),
            n_elements=len(els),
        )

    def as_dict(self):
        """Return a dictionary representation of the Solute."""
        return dict(asdict(self).items())

    # set output of the print() statement
    def __str__(self) -> str:
        return (
            "Species "
            + str(self.formula)
            + " MW="
            + str(self.mw)
            + " Formal Charge="
            + str(self.charge)
            + " Amount= "
            + str(self.moles)
        )



"""
Tests of pyEQL.utils module

"""

from iapws import IAPWS95, IAPWS97
from pytest import raises

from pyEQL import ureg
from pyEQL.utils import FormulaDict, create_water_substance, format_solutes_dict, standardize_formula


def test_standardize_formula():
    """
    Test formula standardization
    """
    assert standardize_formula("H2O") == "H2O(aq)"
    assert standardize_formula("Na+") == "Na[+1]"
    assert standardize_formula("Na[+]") == "Na[+1]"
    assert standardize_formula("SO4--") == "SO4[-2]"
    assert standardize_formula("Mg+2") == "Mg[+2]"
    assert standardize_formula("O2") == "O2(aq)"
    assert standardize_formula("NH4+") == "NH4[+1]"
    assert standardize_formula("NH3") == "NH3(aq)"
    assert standardize_formula("HPO4--") == "HPO4[-2]"
    assert standardize_formula("H2PO4-") == "H2PO4[-1]"
    assert standardize_formula("SCN-") == "SCN[-1]"
    assert standardize_formula("I3-") == "I3[-1]"
    assert standardize_formula("N3-") == "N3[-1]"
    assert standardize_formula("P3-") == "P3[-1]"
    assert standardize_formula("HCOO-") == "HCO2[-1]"
    assert standardize_formula("CO2-1") == "C2O4[-2]"
    assert standardize_formula("C2O4--") == "C2O4[-2]"
    assert standardize_formula("H3PO4") == "H3PO4(aq)"
    assert standardize_formula("H2SO4") == "H2SO4(aq)"
    assert standardize_formula("HClO4") == "HClO4(aq)"
    assert standardize_formula("CF3SO3-") == "CF3SO3[-1]"
    # superscripts, subscripts, and permuted sign/charge number
    assert standardize_formula("PO₄³⁻") == "PO4[-3]"
    assert standardize_formula("Co²⁺") == "Co[+2]"
    # haloacetic acids
    assert standardize_formula("CCl3COO-") == "CCl3COO[-1]"
    assert standardize_formula("CF3COO-") == "CF3COO[-1]"
    assert standardize_formula("CI3COO-") == "CI3COO[-1]"
    assert standardize_formula("CBr3COO-") == "CBr3COO[-1]"
    # Cl+F
    assert standardize_formula("CCl2FCOO-") == "CFCl2COO[-1]"
    assert standardize_formula("CClF2COO-") == "CF2ClCOO[-1]"
    # Cl+I
    assert standardize_formula("CCl2ICOO-") == "CICl2COO[-1]"
    assert standardize_formula("CClI2COO-") == "CI2ClCOO[-1]"
    # Cl+Br
    assert standardize_formula("CBrCl2COO-") == "CBrCl2COO[-1]"
    assert standardize_formula("CBr2ClCOO-") == "CBr2ClCOO[-1]"


def test_formula_dict():
    """
    Make sure flexible key getting/setting works
    """
    d = FormulaDict({"Na+": 0.5, "Cl-1": 1.5, "Mg++": 0.5})
    # assert d == {"Na[+1]": 0.5, "Cl[-1]": 1.5, "Mg[+2]": 0.5}
    # sanitizing of keys should happen automagically
    # assert "Na+" in d
    del d["Mg+2"]
    assert "Mg[+2]" not in d
    assert not d.get("Mg[++]")
    d.update({"Cl-": 0.5})
    print(d, d.data)
    assert d["Cl[-1]"] == 0.5
    assert d.get("Na+") == 0.5
    d.update({"Br-": 2})
    assert d["Br[-]"] == 2


def test_format_solute():
    """
    Test formatting solute dictionaries
    """
    test = {"Na+": 0.5, "Cl-": 0.5}
    base = {"Na+": "0.5 mol/kg", "Cl-": "0.5 mol/kg"}
    assert format_solutes_dict(test, units="mol/kg") == base

    bad_test = [["Na+", 0.5], ["Cl-", 0.5]]
    error_msg = "solute_dict must be a dictionary. Refer to the doc for proper formatting."
    with raises(TypeError, match=error_msg):
        format_solutes_dict(bad_test, units="mol/kg")


def test_create_water_substance():
    assert isinstance(create_water_substance(300, 0.1), IAPWS97)
    assert isinstance(create_water_substance(ureg.Quantity(-15, "degC"), 0.1), IAPWS95)



"""
pyEQL activity correction library.

This file contains functions for computing molal-scale activity coefficients
of ions and salts in aqueous solution.

Individual functions for activity coefficients are defined here so that they
can be used independently of a pyEQL solution object. Normally, these functions
are called from within the get_activity_coefficient method of the Solution class.

:copyright: 2013-2024 by Ryan S. Kingsbury
:license: LGPL, see LICENSE for more details.

"""

import logging

import numpy as np
from pint import Quantity

from pyEQL import ureg
from pyEQL.utils import create_water_substance

logger = logging.getLogger(f"pyEQL.{__name__}")


def _debye_parameter_B(temperature: str = "25 degC") -> Quantity:
    r"""
    Return the constant B used in the extended Debye-Huckel equation.

    Args:
        temperature: The temperature of the solution at which to calculate the constant.
            Defaults to '25 degC'.

    Returns:
        The parameter B for use in extended Debye-Huckel equation (base e). For base 10,
        divide the resulting value by 2.303. Note that A is often given in base 10 terms
        in older textbooks and reference material (0.3281 at 25 degC).

    Notes:
        The parameter B is equal to:

        .. math::

            B = \bigg( \frac{2 N_A \rho_w e^2}{\epsilon_o \epsilon_r k T} \bigg) ^ {\frac{1}{2}}

    References:
        Bockris and Reddy. /Modern Electrochemistry/, vol 1. Plenum/Rosetta, 1977, p.210.

        Archer, Donald G. and Wang, Peiming. "The Dielectric Constant of Water \
        and Debye-Huckel Limiting Law Slopes." /J. Phys. Chem. Ref. Data/ 19(2), 1990.

        https://chem.libretexts.org/Bookshelves/Physical_and_Theoretical_Chemistry_Textbook_Maps/Physical_Chemistry_(LibreTexts)/25%3A_Solutions_II_-_Nonvolatile_Solutes/25.06%3A_The_Debye-Huckel_Theory

        https://en.wikipedia.org/wiki/Debye%E2%80%93H%C3%BCckel_equation
    """
    T = ureg.Quantity(temperature)
    water_substance = create_water_substance(
        T,
        ureg.Quantity(1, "atm"),
    )

    param_B = (
        2
        * ureg.N_A
        * ureg.Quantity(water_substance.rho, "g/L")
        * ureg.elementary_charge**2
        / (ureg.epsilon_0 * water_substance.epsilon * ureg.boltzmann_constant * T)
    ) ** 0.5
    return param_B.to_base_units()


def _debye_parameter_activity(temperature: str = "25 degC") -> "Quantity":
    r"""
    Return the constant A for use in the Debye-Huckel limiting law (base e).

    Args:
        temperature: The temperature of the solution at which to calculate the constant.
            Defaults to '25 degC'.

    Returns:
        The parameter A for use in the Debye-Huckel limiting law (base e). For base 10,
        divide the resulting value by 2.303. Note that A is often given in base 10 terms
        in older textbooks and reference material (0.509 at 25 degC).

    Notes:
        The parameter A is equal to:

        ..  math::

            A^{\gamma} = \frac{e^3 \big( 2 \pi N_A \rho \big)^{0.5}}{(4 \pi \epsilon_o \epsilon_r k T)^{1.5}}

        Note that this equation returns the parameter value that can be used to calculate
        the natural logarithm of the activity coefficient. For base 10, divide the
        value returned by 2.303.

    References:
        Archer, Donald G. and Wang, Peiming. "The Dielectric Constant of Water \
        and Debye-Huckel Limiting Law Slopes." /J. Phys. Chem. Ref. Data/ 19(2), 1990.

        https://chem.libretexts.org/Bookshelves/Physical_and_Theoretical_Chemistry_Textbook_Maps/Physical_Chemistry_(LibreTexts)/25%3A_Solutions_II_-_Nonvolatile_Solutes/25.06%3A_The_Debye-Huckel_Theory

        https://en.wikipedia.org/wiki/Debye%E2%80%93H%C3%BCckel_equation

    See Also:
        :func:`_debye_parameter_osmotic`

    """
    T = ureg.Quantity(temperature)
    water_substance = create_water_substance(
        T,
        ureg.Quantity(1, "atm"),
    )

    debyeparam = (
        ureg.elementary_charge**3
        * (2 * np.pi * ureg.N_A * ureg.Quantity(water_substance.rho, "g/L")) ** 0.5
        / (4 * np.pi * ureg.epsilon_0 * water_substance.epsilon * ureg.boltzmann_constant * T) ** 1.5
    )

    logger.debug(rf"Computed Debye-Huckel Limiting Law Constant A^{{\gamma}} = {debyeparam} at {temperature}")
    return debyeparam.to("kg ** 0.5 / mol ** 0.5")


def _debye_parameter_osmotic(temperature="25 degC"):
    r"""
    Return the constant A_phi for use in calculating the osmotic coefficient according to Debye-Huckel theory.

    Args:
        temperature: String representing the temperature of the solution. Defaults to '25 degC' if not specified.

    Notes:
        Not to be confused with the Debye-Huckel constant used for activity coefficients in the limiting law.
        Takes the value 0.392 at 25 C.
        This constant is calculated according to: [kim]_ [arch]_

        .. math:: A^{\phi} = {1 \over 3} A^{\gamma}

    References:
        .. [kim] Kim, Hee-Talk and Frederick, William Jr, 1988. "Evaluation of Pitzer Ion Interaction Parameters of Aqueous Electrolytes at 25 C. 1. Single Salt Parameters,"
        *J. Chemical Engineering Data* 33, pp.177-184.

        .. [arch] Archer, Donald G. and Wang, Peiming. "The Dielectric Constant of Water \
        and Debye-Huckel Limiting Law Slopes." /J. Phys. Chem. Ref. Data/ 19(2), 1990.

    Examples:
        >>> _debye_parameter_osmotic() #doctest: +ELLIPSIS
        0.3916...

    See Also:
        :func:`_debye_parameter_activity`

    """
    output = 1 / 3 * _debye_parameter_activity(temperature)
    logger.debug(f"Computed Debye-Huckel Limiting slope for osmotic coefficient A^phi = {output} at {temperature}")
    return output.to("kg ** 0.5 /mol ** 0.5")


def _debye_parameter_volume(temperature="25 degC"):
    r"""
    Return the constant A_V, the Debye-Huckel limiting slope for apparent
    molar volume.

    Args:
        temperature: String representing the temperature of the solution. Defaults to '25 degC' if not specified.

    Notes:
        Takes the value 1.8305 cm ** 3 * kg ** 0.5 /  mol ** 1.5 at 25 C.
        This constant is calculated according to: [1]_

        .. math:: A_V = -2 A_{\phi} R T \big[ \frac{3}{\epsilon} \frac{\partial \epsilon}{\partial p} \
         - \frac{1}{\rho}\frac{\partial \rho}{\partial p} \big]

        Notes: at this time, the term in brackets (containing the partial derivatives) is approximate.
        These approximations give the correct value of the slope at 25 degC and
        produce estimates with less than 10% error between 0 and 60 degC.

        The derivative of epsilon with respect to pressure is assumed constant (for atmospheric pressure)
        at -0.01275 1/MPa. Note that the negative sign does not make sense in light
        of real data, but is required to give the correct result.

        The second term is equivalent to the inverse of the bulk modulus of water, which
        is taken to be 2.2 GPa. [2]_

    References:
        .. [1] Archer, Donald G. and Wang, Peiming. "The Dielectric Constant of Water and Debye-Huckel Limiting
               Law Slopes." J. Phys. Chem. Ref. Data/ 19(2), 1990.

        .. [2] http://hyperphysics.phy-astr.gsu.edu/hbase/permot3.html

    See Also:
        :func:`_debye_parameter_osmotic`

    """
    T = ureg.Quantity(temperature)
    water_substance = create_water_substance(
        T,
        ureg.Quantity(1, "atm"),
    )

    # TODO - add partial derivatives to calculation
    epsilon = water_substance.epsilon
    dedp = ureg.Quantity(-0.01275, "1/MPa")
    result = (
        -2 * _debye_parameter_osmotic(temperature) * ureg.R * T * (3 / epsilon * dedp - 1 / ureg.Quantity(2.2, "GPa"))
    )
    # result = ureg.Quantity('1.898 cm ** 3 * kg ** 0.5 /  mol ** 1.5')

    if T.to("degC").magnitude != 25:
        logger.warning("Debye-Huckel limiting slope for volume is approximate when T is not equal to 25 degC")

    logger.debug(f"Computed Debye-Huckel Limiting Slope for volume A^V = {result} at {temperature}")

    return result.to("cm ** 3 * kg ** 0.5 /  mol ** 1.5")


def get_activity_coefficient_debyehuckel(ionic_strength, z=1, temperature="25 degC"):
    r"""
    Return the activity coefficient of solute in the parent solution according to the Debye-Huckel limiting law.

    Args:
        z (int, optional): The charge on the solute, including sign. Defaults to +1 if not specified.
        ionic_strength (Quantity): The ionic strength of the parent solution, mol/kg.
        temperature (str, Quantity, optional): String representing the temperature of the solution. Defaults to '25 degC' if not specified.

    Returns:
        Quantity: The mean molal (mol/kg) scale ionic activity coefficient of solute, dimensionless.

    Notes:
        Activity coefficient is calculated according to:

        .. math:: \ln \gamma = A^{\gamma} z_i^2 \sqrt I

        Valid only for I < 0.005

    See Also:
        :func:`_debye_parameter_activity`
        :func:`get_activity_coefficient_davies`
        :func:`get_activity_coefficient_guntelberg`

    References:
        Stumm, Werner and Morgan, James J. Aquatic Chemistry, 3rd ed,
               pp 103. Wiley Interscience, 1996.
    """
    # check if this method is valid for the given ionic strength
    if not ionic_strength.magnitude <= 0.005:
        logger.warning("Ionic strength exceeds valid range of the Debye-Huckel limiting law")

    log_f = -_debye_parameter_activity(temperature) * z**2 * ionic_strength**0.5

    return np.exp(log_f) * ureg.Quantity(1, "dimensionless")


def get_activity_coefficient_guntelberg(ionic_strength, z=1, temperature="25 degC"):
    r"""
    Return the activity coefficient of solute in the parent solution according to the Guntelberg approximation.

    Args:
        z (int, optional): The charge on the solute, including sign. Defaults to +1 if not specified.
        ionic_strength (Quantity): The ionic strength of the parent solution, mol/kg.
        temperature (str, Quantity, optional): String representing the temperature of the solution. Defaults to '25 degC' if not specified.

    Returns:
        Quantity: The mean molal (mol/kg) scale ionic activity coefficient of solute, dimensionless.

    Notes:
        Activity coefficient is calculated according to:

        .. math:: \ln \gamma = A^{\gamma} z_i^2 {\sqrt I \over (1 + \sqrt I)}

        Valid for I < 0.1

    See Also:
        :func:`_debye_parameter_activity`
        :func:`get_activity_coefficient_davies`
        :func:`get_activity_coefficient_debyehuckel`

    References:
        Stumm, Werner and Morgan, James J. Aquatic Chemistry, 3rd ed,
               pp 103. Wiley Interscience, 1996.
    """
    # check if this method is valid for the given ionic strength
    if not ionic_strength.magnitude <= 0.1:
        logger.warning("Ionic strength exceeds valid range of the Guntelberg approximation")

    log_f = -_debye_parameter_activity(temperature) * z**2 * ionic_strength**0.5 / (1 + ionic_strength.magnitude**0.5)

    return np.exp(log_f) * ureg.Quantity(1, "dimensionless")


def get_activity_coefficient_davies(ionic_strength, z=1, temperature="25 degC"):
    r"""
    Return the activity coefficient of solute in the parent solution according to the Davies equation.

    Args:
        ionic_strength (Quantity): The ionic strength of the parent solution, mol/kg.
        z (int, optional): The charge on the solute, including sign. Defaults to +1 if not specified.
        temperature (str, Quantity, optional): String representing the temperature of the solution. Defaults to '25 degC' if not specified.

    Returns:
        Quantity: The mean molal (mol/kg) scale ionic activity coefficient of solute, dimensionless.

    Notes:
        Activity coefficient is calculated according to:

        .. math:: \ln \gamma = A^{\gamma} z_i^2 ({\sqrt I \over (1 + \sqrt I)} + 0.2 I)

        Valid for 0.1 < I < 0.5

    See Also:
        :func:`_debye_parameter_activity`
        :func:`get_activity_coefficient_debyehuckel`
        :func:`get_activity_coefficient_guntelberg`

    References:
        Stumm, Werner and Morgan, James J. Aquatic Chemistry, 3rd ed,
               pp 103. Wiley Interscience, 1996.
    """
    # check if this method is valid for the given ionic strength
    if not ionic_strength.magnitude <= 0.5 and ionic_strength.magnitude >= 0.1:
        logger.warning("Ionic strength exceeds valid range of the Davies equation")

    # the units in this empirical equation don't work out, so we must use magnitudes
    log_f = (
        -_debye_parameter_activity(temperature).magnitude
        * z**2
        * (ionic_strength.magnitude**0.5 / (1 + ionic_strength.magnitude**0.5) - 0.2 * ionic_strength.magnitude)
    )

    return np.exp(log_f) * ureg.Quantity(1, "dimensionless")


def get_activity_coefficient_pitzer(
    ionic_strength,
    molality,
    alpha1,
    alpha2,
    beta0,
    beta1,
    beta2,
    C_phi,
    z_cation,
    z_anion,
    nu_cation,
    nu_anion,
    temperature="25 degC",
    b=1.2,
):
    """
    Return the activity coefficient of solute in the parent solution according to the Pitzer model.

    Args:
        ionic_strength: The ionic strength of the parent solution, mol/kg
        molality: The molal concentration of the parent salt, mol/kg
        alpha1, alpha2: Coefficients for the Pitzer model. This function assigns the coefficients
            proper units of kg ** 0.5 / mol ** 0.5 after they are entered.
        beta0, beta1, beta2, C_phi:  Coefficients for the Pitzer model. These ion-interaction parameters are
            specific to each salt system.
        z_cation, z_anion: The charge on the cation and anion, respectively
        nu_cation, nu_anion: The stoichiometric coefficient of the cation and anion in the salt
        temperature: String representing the temperature of the solution. Defaults to '25 degC' if not specified.
        b: Coefficient. Usually set equal to 1.2 and considered independent of temperature and pressure.
            If provided, this coefficient is assigned proper units of kg ** 0.5 / mol ** 0.5  after entry.

    Returns:
        Quantity
            The mean molal (mol/kg) scale ionic activity coefficient of solute, dimensionless

    Examples:
        >>> get_activity_coefficient_pitzer(0.5*ureg.Quantity('mol/kg'),0.5*ureg.Quantity('mol/kg'),1,0.5,-.0181191983,-.4625822071,.4682,.000246063,1,-1,1,1,b=1.2)
        ￼0.61915...

        >>> get_activity_coefficient_pitzer(5.6153*ureg.Quantity('mol/kg'),5.6153*ureg.Quantity('mol/kg'),3,0.5,0.0369993,0.354664,0.0997513,-0.00171868,1,-1,1,1,b=1.2)
        ￼0.76331...

        Notes: the examples below are for comparison with experimental and modeling data presented in
        the May et al reference below.

        10 mol/kg ammonium nitrate. Estimated result (from graph) = 0.2725

        >>> get_activity_coefficient_pitzer(10*ureg.Quantity('mol/kg'),10*ureg.Quantity('mol/kg'),2,0,-0.01709,0.09198,0,0.000419,1,-1,1,1,b=1.2)
        0.22595 ...

        5 mol/kg ammonium nitrate. Estimated result (from graph) = 0.3011

        >>> get_activity_coefficient_pitzer(5*ureg.Quantity('mol/kg'),5*ureg.Quantity('mol/kg'),2,0,-0.01709,0.09198,0,0.000419,1,-1,1,1,b=1.2)
        0.30249 ...

        18 mol/kg ammonium nitrate. Estimated result (from graph) = 0.1653

        >>> get_activity_coefficient_pitzer(18*ureg.Quantity('mol/kg'),18*ureg.Quantity('mol/kg'),2,0,-0.01709,0.09198,0,0.000419,1,-1,1,1,b=1.2)
        0.16241 ...

    References:
        Scharge, T., Munoz, A.G., and Moog, H.C. (2012). Activity Coefficients of Fission Products in Highly
        Salinary Solutions of Na+, K+, Mg2+, Ca2+, Cl-, and SO42- : Cs+.
        /Journal of Chemical& Engineering Data (57), p. 1637-1647.

        Kim, H., & Jr, W. F. (1988). Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
        Journal of Chemical and Engineering Data, (2), 177-184.

        May, P. M., Rowland, D., Hefter, G., & Königsberger, E. (2011).
        A Generic and Updatable Pitzer Characterization of Aqueous Binary Electrolyte Solutions at 1 bar and 25 °C.
        Journal of Chemical & Engineering Data, 56(12), 5066-5077. doi:10.1021/je2009329

        Beyer, R., & Steiger, M. (2010). Vapor Pressure Measurements of NaHCOO + H 2 O and KHCOO + H 2 O from 278 to 308 K
        and Representation with an Ion Interaction (Pitzer) Model.
        Journal of Chemical & Engineering Data, 55(2), 830-838. doi:10.1021/je900487a

    See Also:
        :func:`_debye_parameter_activity`
        :func:`_pitzer_B_MX`
        :func:`_pitzer_B_phi`
        :func:`_pitzer_log_gamma`
    """
    # assign proper units to alpha1, alpha2, and b
    alpha1 = ureg.Quantity(alpha1, "kg ** 0.5 / mol ** 0.5")
    alpha2 = ureg.Quantity(alpha2, "kg ** 0.5 / mol ** 0.5")
    b = ureg.Quantity(b, "kg ** 0.5 / mol ** 0.5")
    C_phi = ureg.Quantity(C_phi, "kg ** 2 /mol ** 2")

    # assign units appropriate for the activity parameters
    BMX = ureg.Quantity(_pitzer_B_MX(ionic_strength, alpha1, alpha2, beta0, beta1, beta2), "kg/mol")
    Bphi = ureg.Quantity(_pitzer_B_phi(ionic_strength, alpha1, alpha2, beta0, beta1, beta2), "kg/mol")

    loggamma = _pitzer_log_gamma(
        ionic_strength,
        molality,
        BMX,
        Bphi,
        C_phi,
        z_cation,
        z_anion,
        nu_cation,
        nu_anion,
        temperature,
        b,
    )

    return np.exp(loggamma) * ureg.Quantity(1, "dimensionless")


def get_apparent_volume_pitzer(
    ionic_strength,
    molality,
    alpha1,
    alpha2,
    beta0,
    beta1,
    beta2,
    C_phi,
    V_o,
    z_cation,
    z_anion,
    nu_cation,
    nu_anion,
    temperature="25 degC",
    b=1.2,
):
    """
    Return the apparent molar volume of solute in the parent solution according to the Pitzer model.

    Args:
        ionic_strength (Quantity): The ionic strength of the parent solution, mol/kg.
        molality (Quantity): The molal concentration of the parent salt, mol/kg.
        alpha1, alpha2 (number): Coefficients for the Pitzer model. This function assigns the coefficients proper units of kg ** 0.5 / mol ** 0.5 after they are entered.
        beta0, beta1, beta2, C_phi (number): Pitzer coefficients for the apparent molar volume. These ion-interaction parameters are specific to each salt system.
        V_o (number): The V^o Pitzer coefficient for the apparent molar volume.
        z_cation, z_anion (int): The formal charge on the cation and anion, respectively.
        nu_cation, nu_anion (int): The stoichiometric coefficient of the cation and anion in the salt.
        temperature (str, Quantity): String representing the temperature of the solution. Defaults to '25 degC' if not specified.
        b (number, optional): Coefficient. Usually set equal to 1.2 and considered independent of temperature and pressure. If provided, this coefficient is assigned proper units of kg ** 0.5 / mol ** 0.5  after entry.

    Returns:
        Quantity: The apparent molar volume of the solute, cm ** 3 / mol.

    Examples:
        Notes: the example below is for comparison with experimental and modeling data presented in the Krumgalz et al reference below.

        0.25 mol/kg CuSO4. Expected result (from graph) = 0.5 cm ** 3 / mol

        >>> get_apparent_volume_pitzer(1.0*ureg.Quantity('mol/kg'),0.25*ureg.Quantity('mol/kg'),1.4,12,0.001499,-0.008124,0.2203,-0.0002589,-6,2,-2,1,1,b=1.2)
        0.404...

        1.0 mol/kg CuSO4. Expected result (from graph) = 4 cm ** 3 / mol

        >>> get_apparent_volume_pitzer(4.0*ureg.Quantity('mol/kg'),1.0*ureg.Quantity('mol/kg'),1.4,12,0.001499,-0.008124,0.2203,-0.0002589,-6,2,-2,1,1,b=1.2)
        4.424...

        10.0 mol/kg ammonium nitrate. Expected result (from graph) = 50.3 cm ** 3 / mol

        >>> get_apparent_volume_pitzer(10.0*ureg.Quantity('mol/kg'),10.0*ureg.Quantity('mol/kg'),2,0,0.000001742,0.0002926,0,0.000000424,46.9,1,-1,1,1,b=1.2)
        50.286...

        20.0 mol/kg ammonium nitrate. Expected result (from graph) = 51.2 cm ** 3 / mol

        >>> get_apparent_volume_pitzer(20.0*ureg.Quantity('mol/kg'),20.0*ureg.Quantity('mol/kg'),2,0,0.000001742,0.0002926,0,0.000000424,46.9,1,-1,1,1,b=1.2)
        51.145...

        Notes: the examples below are for comparison with experimental and modeling data presented in the Krumgalz et al reference below.

        0.8 mol/kg NaF. Expected result = 0.03

        >>> get_apparent_volume_pitzer(0.8*ureg.Quantity('mol/kg'),0.8*ureg.Quantity('mol/kg'),2,0,0.000024693,0.00003169,0,-0.000004068,-2.426,1,-1,1,1,b=1.2)
        0.22595 ...

    References:
        May, P. M., Rowland, D., Hefter, G., & Königsberger, E. (2011).
        A Generic and Updatable Pitzer Characterization of Aqueous Binary Electrolyte Solutions at 1 bar and 25 °C.
        Journal of Chemical & Engineering Data, 56(12), 5066-5077. doi:10.1021/je2009329

        Krumgalz, Boris S., Pogorelsky, Rita (1996).
        Volumetric Properties of Single Aqueous Electrolytes from Zero to Saturation Concentration at 298.15 K
        Represented by Pitzer's Ion-Interaction Equations.
        Journal of Physical Chemical Reference Data, 25(2), 663-689.

    See Also:
        :func:`_debye_parameter_volume`
        :func:`_pitzer_B_MX`
    """
    # TODO - find a cleaner way to make sure coefficients are assigned the proper units
    # if they aren't, the calculation gives very wrong results
    alpha1 = ureg.Quantity(alpha1, "kg ** 0.5 / mol ** 0.5")
    alpha2 = ureg.Quantity(alpha2, "kg ** 0.5 / mol ** 0.5")
    b = ureg.Quantity(b, "kg ** 0.5 / mol ** 0.5")
    C_phi = ureg.Quantity(C_phi, "kg ** 2 /mol ** 2 / dabar")
    V_o = ureg.Quantity(V_o, "cm ** 3 / mol")

    # assign units appropriate for the volume parameter
    BMX = ureg.Quantity(_pitzer_B_MX(ionic_strength, alpha1, alpha2, beta0, beta1, beta2), "kg /mol/dabar")

    second_term = (
        (nu_cation + nu_anion)
        * abs(z_cation * z_anion)
        * (_debye_parameter_volume(temperature) / 2 / b)
        * np.log(1 + b * ionic_strength**0.5)
    )

    third_term = (
        nu_cation
        * nu_anion
        * ureg.R
        * ureg.Quantity(temperature)
        * (2 * molality * BMX + molality**2 * C_phi * (nu_cation * nu_anion) ** 0.5)
    )

    volume = V_o + second_term + third_term

    return volume.to("cm ** 3 / mol")


def _pitzer_f1(x):
    r"""
    The function of ionic strength used to calculate \beta_MX in the Pitzer ion interaction model.

    .. math:: f(x) = 2 [ 1- (1+x) \exp(-x)] / x ^ 2

    References:
        Scharge, T., Munoz, A.G., and Moog, H.C. (2012). Activity Coefficients of Fission Products in Highly
        Salinary Solutions of Na+, K+, Mg2+, Ca2+, Cl-, and SO42- : Cs+.
        /Journal of Chemical& Engineering Data (57), p. 1637-1647.

        Kim, H., & Jr, W. F. (1988). Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
        Journal of Chemical and Engineering Data, (2), 177-184.
    """
    # return 0 if the input is 0
    if x == 0:
        return 0
    return 2 * (1 - (1 + x) * np.exp(-x)) / x**2


def _pitzer_f2(x):
    r"""
    The function of ionic strength used to calculate \beta_\gamma in the Pitzer ion interaction model.

    .. math:: f(x) = -\frac{2}{x ^ 2} [ 1 - (\frac{1+x+ x^2}{2}) \exp(-x)]

    References:
        Scharge, T., Munoz, A.G., and Moog, H.C. (2012). Activity Coefficients of Fission Products in Highly
        Salinary Solutions of Na+, K+, Mg2+, Ca2+, Cl-, and SO42- : Cs+.
        /Journal of Chemical& Engineering Data (57), p. 1637-1647.

        Kim, H., & Jr, W. F. (1988). Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
        Journal of Chemical and Engineering Data, (2), 177-184.
    """
    # return 0 if the input is 0
    if x == 0:
        return 0
    return -2 * (1 - (1 + x + x**2 / 2) * np.exp(-x)) / x**2


def _pitzer_B_MX(ionic_strength, alpha1, alpha2, beta0, beta1, beta2):
    r"""
    Return the B_MX coefficient for the Pitzer ion interaction model.

    .. math:: B_MX = \beta_0 + \beta_1 f1(\alpha_1 I ^ {0.5}) + \beta_2 f2(\alpha_2 I ^ {0.5})

    Args:
        ionic_strength: The ionic strength of the parent solution, mol/kg
        alpha1, alpha2: Coefficients for the Pitzer model, kg ** 0.5 / mol ** 0.5
        beta0, beta1, beta2: Coefficients for the Pitzer model. These ion-interaction parameters are
            specific to each salt system.

    Returns:
        The B_MX parameter for the Pitzer ion interaction model.

    References:
        Scharge, T., Munoz, A.G., and Moog, H.C. (2012). Activity Coefficients of Fission Products in Highly
        Salinary Solutions of Na+, K+, Mg2+, Ca2+, Cl-, and SO42- : Cs+.
        /Journal of Chemical& Engineering Data (57), p. 1637-1647.

        Kim, H., & Jr, W. F. (1988). Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
        Journal of Chemical and Engineering Data, (2), 177-184.

    See Also:
        :func:`_pitzer_f1`

    """
    coeff = beta0 + beta1 * _pitzer_f1(alpha1 * ionic_strength**0.5) + beta2 * _pitzer_f1(alpha2 * ionic_strength**0.5)
    return coeff.magnitude


# def _pitzer_B_gamma(ionic_strength,alpha1,alpha2,beta1,beta2):
#    '''
#    Return the B^\gamma coefficient for the Pitzer ion interaction model.
#
#    .. math:: B_\gamma = [ \beta_1 f2(\alpha_1 I ^ 0.5) + beta_2 f2(\alpha_2 I^0.5) ] / I
#
#    Parameters
#    ----------
#    ionic_strength: number
#                    The ionic strength of the parent solution, mol/kg
#    alpha1, alpha2: number
#                    Coefficients for the Pitzer model, kg ** 0.5 / mol ** 0.5.
#    beta1, beta2: number
#                    Coefficients for the Pitzer model. These ion-interaction parameters are
#                    specific to each salt system.
#
#    Returns
#    -------
#    float
#            The B^gamma parameter for the Pitzer ion interaction model.
#
#    References
#    ----------
#    Scharge, T., Munoz, A.G., and Moog, H.C. (2012). Activity Coefficients of Fission Products in Highly
#    Salinary Solutions of Na+, K+, Mg2+, Ca2+, Cl-, and SO42- : Cs+.
#    /Journal of Chemical& Engineering Data (57), p. 1637-1647.
#
#    Kim, H., & Jr, W. F. (1988). Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
#    Journal of Chemical and Engineering Data, (2), 177-184.
#
#    See Also
#    --------
#    _pitzer_f2
#
#    '''
#    coeff = (beta1 * _pitzer_f2(alpha1 * ionic_strength ** 0.5) + beta2 * _pitzer_f2(alpha2 * ionic_strength ** 0.5)) / ionic_strength
#    return coeff * ureg.Quantity('kg/mol')


def _pitzer_B_phi(ionic_strength, alpha1, alpha2, beta0, beta1, beta2):
    r"""
    Returns the B^\Phi coefficient for the Pitzer ion interaction model.

    This function calculates the B^\Phi coefficient using the formula:

    .. math:: B^\Phi = \beta_0 + \beta1 \exp(-\alpha_1 I ^{0.5}) + \beta_2 \exp(-\alpha_2 I ^ {0.5})

    or

    .. math:: B^\Phi = B^\gamma - B_{MX}

    Args:
        ionic_strength: The ionic strength of the parent solution, mol/kg.
        alpha1, alpha2: Coefficients for the Pitzer model, kg ** 0.5 / mol ** 0.5.
        beta0, beta1, beta2: Coefficients for the Pitzer model. These ion-interaction parameters are
            specific to each salt system.

    Returns:
        float: The B^Phi parameter for the Pitzer ion interaction model.

    References:
        Scharge, T., Munoz, A.G., and Moog, H.C. (2012). Activity Coefficients of Fission Products in Highly
        Salinary Solutions of Na+, K+, Mg2+, Ca2+, Cl-, and SO42- : Cs+.
        /Journal of Chemical& Engineering Data (57), p. 1637-1647.

        Kim, H., & Jr, W. F. (1988). Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
        Journal of Chemical and Engineering Data, (2), 177-184.

        Beyer, R., & Steiger, M. (2010). Vapor Pressure Measurements of NaHCOO + H 2 O and KHCOO + H 2 O from 278 to 308 K
        and Representation with an Ion Interaction (Pitzer) Model.
        Journal of Chemical & Engineering Data, 55(2), 830-838. doi:10.1021/je900487a
    """
    import math

    # TODO - for some reason this specific method requires the use of math.exp rather than np.exp. Using np.exp raises
    # a dimensionalityerror.
    return beta0 + beta1 * math.exp(-alpha1 * ionic_strength**0.5) + beta2 * math.exp(-alpha2 * ionic_strength**0.5)


# def _pitzer_C_MX(C_phi,z_cation,z_anion):
#    '''
#    Return the C^\Phi coefficient for the Pitzer ion interaction model.
#
#    .. math:: C_MX = C^\Phi / 2 \sqrt( \abs(z_+ z_-))
#
#    Parameters
#    ----------
#    C_phi: number
#                    The C_phi parameter for the Pitzer ion interaction model.
#    z_cation, z_anion: int
#                    The formal charge on the cation and anion, respectively
#
#    Returns
#    -------
#    float
#            The C_MX parameter for the Pitzer ion interaction model.
#
#    References
#    ----------
#    Kim, H., & Jr, W. F. (1988).
#    Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
#    Journal of Chemical and Engineering Data, (2), 177-184.
#
#    May, P. M., Rowland, D., Hefter, G., & Königsberger, E. (2011).
#    A Generic and Updatable Pitzer Characterization of Aqueous Binary Electrolyte Solutions at 1 bar and 25 °C.
#    Journal of Chemical & Engineering Data, 56(12), 5066-5077. doi:10.1021/je2009329
#    '''
#
#    coeff = C_phi / ( 2 * abs(z_cation * z_anion) ** 0.5 )
#    return coeff * ureg.Quantity('kg ** 2 /mol ** 2')


def _pitzer_log_gamma(
    ionic_strength,
    molality,
    B_MX,
    B_phi,
    C_phi,
    z_cation,
    z_anion,
    nu_cation,
    nu_anion,
    temperature="25 degC",
    b=ureg.Quantity(1.2, "kg**0.5/mol**0.5"),
):
    r"""
    Returns the natural logarithm of the binary activity coefficient calculated by the Pitzer
    ion interaction model.

    .. math::

        \ln \gamma_{MX} = -|z_+ z_-| A^{Phi} ( \frac{I ^ {0.5}}{(1 + b I ^ {0.5})} + \frac{2}{b} \ln{(1 + b I ^ {0.5})} )+ \frac{m (2 \nu_+ \nu_-)}{(\nu_+ + \nu_-)} (B_{MX} + B_{MX}^\Phi) + \frac{m^2(3 (\nu_+ \nu_-)^{1.5}}{(\nu_+ + \nu_-))} C_{MX}^\Phi

    Args:
        ionic_strength (Quantity): The ionic strength of the parent solution, mol/kg.
        molality (Quantity): The concentration of the salt, mol/kg.
        B_MX, B_phi, C_phi (Quantity): Calculated parameters for the Pitzer ion interaction model.
        z_cation, z_anion (int): The formal charge on the cation and anion, respectively.
        nu_cation, nu_anion (int): The stoichiometric coefficient of the cation and anion in the salt.
        temperature (str, Quantity): String representing the temperature of the solution. Defaults to '25 degC' if not specified.
        b (number, optional): Coefficient. Usually set equal to 1.2 kg ** 0.5 / mol ** 0.5 and considered independent of temperature and pressure.

    Returns:
        float: The natural logarithm of the binary activity coefficient calculated by the Pitzer ion interaction model.

    References:
        Kim, H., & Jr, W. F. (1988). Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
        Journal of Chemical and Engineering Data, (2), 177-184.

        May, P. M., Rowland, D., Hefter, G., & Königsberger, E. (2011). A Generic and Updatable Pitzer Characterization of Aqueous Binary Electrolyte Solutions at 1 bar and 25 °C.
        Journal of Chemical & Engineering Data, 56(12), 5066-5077. doi:10.1021/je2009329
    """
    first_term = (
        -1
        * abs(z_cation * z_anion)
        * _debye_parameter_osmotic(temperature)
        * (ionic_strength**0.5 / (1 + b * ionic_strength**0.5) + 2 / b * np.log(1 + b * ionic_strength**0.5))
    )
    second_term = 2 * molality * nu_cation * nu_anion / (nu_cation + nu_anion) * (B_MX + B_phi)
    third_term = 3 * molality**2 * (nu_cation * nu_anion) ** 1.5 / (nu_cation + nu_anion) * C_phi

    return first_term + second_term + third_term


def get_osmotic_coefficient_pitzer(
    ionic_strength,
    molality,
    alpha1,
    alpha2,
    beta0,
    beta1,
    beta2,
    C_phi,
    z_cation,
    z_anion,
    nu_cation,
    nu_anion,
    temperature="25 degC",
    b=1.2,
):
    """
    Return the osmotic coefficient of water in an electrolyte solution according to the Pitzer model.

    Args:
        ionic_strength (Quantity): The ionic strength of the parent solution, mol/kg.
        molality (Quantity): The molal concentration of the parent salt, mol/kg.
        alpha1, alpha2 (number): Coefficients for the Pitzer model. This function assigns the coefficients proper units of kg ** 0.5 / mol ** 0.5 after they are entered.
        beta0, beta1, beta2, C_phi: Coefficients for the Pitzer model. These ion-interaction parameters are specific to each salt system.
        z_cation, z_anion (int): The formal charge on the cation and anion, respectively.
        nu_cation, nu_anion (int): The stoichiometric coefficient of the cation and anion in the salt.
        temperature (str, Quantity): String representing the temperature of the solution. Defaults to '25 degC' if not specified.
        b (number, optional): Coefficient. Usually set equal to 1.2 and considered independent of temperature and pressure. If provided, this coefficient is assigned proper units of kg ** 0.5 / mol ** 0.5  after entry.

    Returns:
        Quantity: The osmotic coefficient of water, dimensionless.

    Examples:
        Experimental value according to Beyer and Stieger reference is 1.3550

        >>> get_osmotic_coefficient_pitzer(10.175*ureg.Quantity('mol/kg'),10.175*ureg.Quantity('mol/kg'),1,0.5,-.0181191983,-.4625822071,.4682,.000246063,1,-1,1,1,b=1.2)
        1.3552 ...

        Experimental value according to Beyer and Stieger reference is 1.084

        >>> get_osmotic_coefficient_pitzer(5.6153*ureg.Quantity('mol/kg'),5.6153*ureg.Quantity('mol/kg'),3,0.5,0.0369993,0.354664,0.0997513,-0.00171868,1,-1,1,1,b=1.2)
        1.0850 ...

        Notes: the examples below are for comparison with experimental and modeling data presented in the May et al reference below.

        10 mol/kg ammonium nitrate. Estimated result (from graph) = 0.62

        >>> get_osmotic_coefficient_pitzer(10*ureg.Quantity('mol/kg'),10*ureg.Quantity('mol/kg'),2,0,-0.01709,0.09198,0,0.000419,1,-1,1,1,b=1.2)
        0.6143 ...

        5 mol/kg ammonium nitrate. Estimated result (from graph) = 0.7

        >>> get_osmotic_coefficient_pitzer(5*ureg.Quantity('mol/kg'),5*ureg.Quantity('mol/kg'),2,0,-0.01709,0.09198,0,0.000419,1,-1,1,1,b=1.2)
        0.6925 ...

        18 mol/kg ammonium nitrate. Estimated result (from graph) = 0.555

        >>> get_osmotic_coefficient_pitzer(18*ureg.Quantity('mol/kg'),18*ureg.Quantity('mol/kg'),2,0,-0.01709,0.09198,0,0.000419,1,-1,1,1,b=1.2)
        0.5556 ...

    References:
        Scharge, T., Munoz, A.G., and Moog, H.C. (2012). Activity Coefficients of Fission Products in Highly Salinary Solutions of Na+, K+, Mg2+, Ca2+, Cl-, and SO42- : Cs+.
        /Journal of Chemical& Engineering Data (57), p. 1637-1647.

        Kim, H., & Jr, W. F. (1988). Evaluation of Pitzer ion interaction parameters of aqueous electrolytes at 25 degree C. 1. Single salt parameters.
        Journal of Chemical and Engineering Data, (2), 177-184.

        May, P. M., Rowland, D., Hefter, G., & Königsberger, E. (2011). A Generic and Updatable Pitzer Characterization of Aqueous Binary Electrolyte Solutions at 1 bar and 25 °C.
        Journal of Chemical & Engineering Data, 56(12), 5066-5077. doi:10.1021/je2009329

        Beyer, R., & Steiger, M. (2010). Vapor Pressure Measurements of NaHCOO + H 2 O and KHCOO + H 2 O from 278 to 308 K and Representation with an Ion Interaction (Pitzer) Model.
        Journal of Chemical & Engineering Data, 55(2), 830-838. doi:10.1021/je900487a

    See Also:
        :func:`_debye_parameter_activity`
        :func:`_pitzer_B_MX`
        :func:`_pitzer_B_phi`
        :func:`_pitzer_log_gamma`
    """
    # assign proper units to alpha1, alpha2, and b
    alpha1 = ureg.Quantity(alpha1, "kg ** 0.5 / mol ** 0.5")
    alpha2 = ureg.Quantity(alpha2, "kg ** 0.5 / mol ** 0.5")
    b = ureg.Quantity(b, "kg ** 0.5 / mol ** 0.5")
    C_phi = ureg.Quantity(C_phi, "kg ** 2 /mol ** 2")
    B_phi = ureg.Quantity(_pitzer_B_phi(ionic_strength, alpha1, alpha2, beta0, beta1, beta2), "kg/mol")

    first_term = 1 - _debye_parameter_osmotic(temperature) * abs(z_cation * z_anion) * ionic_strength**0.5 / (
        1 + b * ionic_strength**0.5
    )
    second_term = molality * 2 * nu_cation * nu_anion / (nu_cation + nu_anion) * B_phi
    third_term = molality**2 * (2 * (nu_cation * nu_anion) ** 1.5 / (nu_cation + nu_anion)) * C_phi

    return first_term + second_term + third_term



"""
pyEQL volume and concentration methods test suite
=================================================

This file contains tests for the volume and concentration-related methods
used by pyEQL's Solution class
"""

import numpy as np
import pytest

import pyEQL


@pytest.fixture()
def s1():
    return pyEQL.Solution(volume="2 L")


@pytest.fixture()
def s2():
    return pyEQL.Solution([["Na+", "4 mol/L"], ["Cl-", "4 mol/L"]], volume="2 L")


@pytest.fixture()
def s3():
    return pyEQL.Solution([["Na+", "4 mol/kg"], ["Cl-", "4 mol/kg"]], volume="2 L")


@pytest.fixture()
def s4():
    return pyEQL.Solution([["Na+", "8 mol"], ["Cl-", "8 mol"]], volume="2 L")


class Test_solute_addition:
    """
    test behavior of various methods for adding solutes to a solution
    -----------------------------------------------------------------

    """

    def test_set_amount_1(self, s2):
        """
        Tests for set_amount() method
        """
        # If the concentration of a solute is directly set with a substance / volume
        # unit, the volume should not change
        s2.set_amount("Na+", "5 mol/L")
        s2.set_amount("Cl-", "5 mol/L")
        assert s2.volume.to("L").magnitude == 2

    def test_set_amount_2(self, s2):
        # If the concentration of a solute is directly set with a substance / volume
        # unit, the water mass should be reduced
        original = s2.solvent_mass.to("kg").magnitude
        s2.set_amount("Na+", "5 mol/L")
        s2.set_amount("Cl-", "5 mol/L")
        assert s2.solvent_mass.to("kg").magnitude < original

    def test_set_amount_3(self, s2):
        # If the concentration of a solute is directly set with a substance / volume
        # unit, the resulting concentration should be exactly what was specified
        s2.set_amount("Na+", "5 mol/L")
        s2.set_amount("Cl-", "5 mol/L")
        assert np.allclose(s2.get_amount("Na+", "mol/L").magnitude, 5)

    def test_set_amount_4(self, s2):
        # If the concentration of a solute is directly set with a substance / mass
        # unit, the volume should increase
        original = s2.volume.to("L").magnitude
        s2.set_amount("Na+", "5 mol/kg")
        s2.set_amount("Cl-", "5 mol/kg")
        assert s2.volume.to("L").magnitude > original

    def test_set_amount_5(self, s2):
        # If the concentration of a solute is directly set with a substance / mass
        # unit, the water mass should not change
        original = s2.solvent_mass.to("kg").magnitude
        s2.set_amount("Na+", "5 mol/kg")
        s2.set_amount("Cl-", "5 mol/kg")
        assert np.allclose(s2.solvent_mass.to("kg").magnitude, original)

    def test_set_amount_6(self, s2):
        # If the concentration of a solute is directly set with a substance / mass
        # unit, the resulting concentration should be exactly what was specified
        s2.set_amount("Na+", "5 mol/kg")
        s2.set_amount("Cl-", "5 mol/kg")
        assert np.allclose(s2.get_amount("Na+", "mol/kg").magnitude, 5)

    def test_set_amount_7(self, s2):
        # If the concentration of a solute is directly set with a substance
        # unit, the volume should increase
        original = s2.volume.to("L").magnitude
        s2.set_amount("Na+", "10 mol")
        s2.set_amount("Cl-", "10 mol")
        assert s2.volume.to("L").magnitude > original

    def test_set_amount_8(self, s2):
        # If the concentration of a solute is directly set with a substance
        # unit, the water mass should not change
        original = s2.solvent_mass.to("kg").magnitude
        s2.set_amount("Na+", "10 mol")
        s2.set_amount("Cl-", "10 mol")
        assert np.allclose(s2.solvent_mass.to("kg").magnitude, original)

    def test_set_amount_9(self, s2):
        # If the concentration of a solute is directly set with a substance / mass
        # unit, the resulting concentration should be exactly what was specified
        s2.set_amount("Na+", "10 mol")
        s2.set_amount("Cl-", "10 mol")
        assert np.allclose(s2.get_amount("Na+", "mol").magnitude, 10)

    def test_add_amount_1(self, s2):
        """
        Tests for add_amount() method
        """
        # substance / volume units
        # If the concentration of a solute is directly increased with a substance / volume
        # unit, the volume should not change
        s2.add_amount("Na+", "1 mol/L")
        s2.add_amount("Cl-", "1 mol/L")
        assert np.allclose(s2.volume.to("L").magnitude, 2)

    def test_add_amount_2(self, s2):
        # If the concentration of a solute is directly increased with a substance / volume
        # unit, the water mass should be reduced
        original = s2.solvent_mass.to("kg").magnitude
        s2.add_amount("Na+", "1 mol/L")
        s2.add_amount("Cl-", "1 mol/L")
        assert s2.solvent_mass.to("kg").magnitude < original

    def test_add_amount_3(self, s2):
        # If the concentration of a solute is directly increased with a substance / volume
        # unit, the resulting concentration should be exactly what was specified
        s2.add_amount("Na+", "1 mol/L")
        s2.add_amount("Cl-", "1 mol/L")
        assert np.allclose(s2.get_amount("Na+", "mol/L").magnitude, 5)

    def test_add_amount_4(self, s3):
        # substance / mass units

        # If the concentration of a solute is directly increased with a substance / mass
        # unit, the volume should increase
        original = s3.volume.to("L").magnitude
        s3.add_amount("Na+", "1 mol/kg")
        s3.add_amount("Cl-", "1 mol/kg")
        assert s3.volume.to("L").magnitude > original

    def test_add_amount_5(self, s3):
        # If the concentration of a solute is directly increased with a substance / mass
        # unit, the water mass should not change
        original = s3.solvent_mass.to("kg").magnitude
        V_orig = s3.volume.to("L").magnitude
        s3.add_amount("Na+", "1 mol/kg")
        s3.add_amount("Cl-", "1 mol/kg")
        assert np.allclose(s3.solvent_mass.to("kg").magnitude, original)
        assert s3.volume.to("L").magnitude > V_orig
        # pH will be slightly higher than 7 b/c the addition of solute caused the
        # solution volume to increase, so lower mol/L = higher pH
        assert np.isclose(s3.pH, 7.0, atol=0.05)
        assert np.isclose(s3.pE, 8.5)

    def test_add_amount_12(self, s3):
        # If the concentration of a solute is directly increased with a substance / mass
        # unit, the resulting concentration should be exactly what was specified
        s3.add_amount("Na+", "1 mol/kg")
        s3.add_amount("Cl-", "1 mol/kg")
        assert np.allclose(s3.get_amount("Na+", "mol/kg").magnitude, 5)

    def test_add_amount_6(self, s2):
        # substance units

        # If the concentration of a solute is directly increased with a substance
        # unit, the volume should increase
        original = s2.volume.to("L").magnitude
        s2.add_amount("Na+", "2 mol")
        s2.add_amount("Cl-", "2 mol")
        assert s2.volume.to("L").magnitude > original

    def test_add_amount_7(self, s2):
        # If the concentration of a solute is directly increased with a substance
        # unit, the water mass should not change
        original = s2.solvent_mass.to("kg").magnitude
        s2.add_amount("Na+", "2 mol")
        s2.add_amount("Cl-", "2 mol")
        assert np.allclose(s2.solvent_mass.to("kg").magnitude, original)

    def test_add_amount_8(self, s2):
        # If the concentration of a solute is directly increased with a substance
        # unit, the resulting concentration should be exactly what was specified
        s2.add_amount("Na+", "2 mol")
        s2.add_amount("Cl-", "2 mol")
        assert np.allclose(s2.get_amount("Na+", "mol").magnitude, 10)

    def test_add_amount_9(self, s2):
        # negative substance units
        # If the concentration of a solute is directly decreased with a substance
        # unit, the volume should decrease
        original = s2.volume.to("L").magnitude
        s2.add_amount("Na+", "-2 mol")
        s2.add_amount("Cl-", "-2 mol")
        assert s2.volume.to("L").magnitude < original

    def test_add_amount_10(self, s2):
        # If the concentration of a solute is directly changed with a substance
        # unit, the water mass should not change
        original = s2.solvent_mass.to("kg").magnitude
        s2.add_amount("Na+", "-2 mol")
        s2.add_amount("Cl-", "-2 mol")
        assert np.allclose(s2.solvent_mass.to("kg").magnitude, original)

    def test_add_amount_11(self, s2):
        # If the concentration of a solute is directly changed with a substance
        # unit, the resulting concentration should be exactly what was specified
        s2.add_amount("Na+", "-2 mol")
        s2.add_amount("Cl-", "-2 mol")
        assert np.allclose(s2.get_amount("Na+", "mol").magnitude, 6)

    def test_add_amount_12(self, s2):
        # test behavior when the solute is initially absent from the solution
        s2.add_amount("Ca+2", "1 mol")
        s2.add_amount("Br-", "1 mol")
        assert np.allclose(s2.get_amount("Ca+2", "mol/L").magnitude, 0.5, atol=0.002)
        assert np.allclose(s2.get_amount("Br-", "mol/L").magnitude, 0.5, atol=0.002)


class Test_get_amount:
    """
    test the get_amount() method on a 1 mol/L NaCl solution
    ----------------------------
    1 mol NaCl / L = 58.44 g/L
    Na+ = 22.98977 g/mol

    """

    def test_get_amount_nacl(self):
        # create the 1 M NaCl solution
        s1 = pyEQL.Solution([["Na+", "1 mol/L"], ["Cl-", "1 mol/L"]])

        # get_amount() - mol/L
        assert np.isclose(s1.get_amount("Na+", "mol/L").magnitude, 1)

        # get_amount() - mol/kg
        assert np.isclose(s1.get_amount("Na+", "mol/kg").magnitude, 1.02181221888)

        # get_amount() - g/L
        assert np.isclose(s1.get_amount("Na+", "g/L").magnitude, 22.98977)

        # get_amount() - mg
        assert np.isclose(s1.get_amount("Na+", "mg").magnitude, 22989.77)

        # get_amount() - mol
        assert np.isclose(s1.get_amount("Na+", "mol").magnitude, 1)

        # get_amount() - fraction
        assert np.isclose(s1.get_amount("Na+", "fraction"), 0.01775457254)

        # get_amount() - count
        assert np.isclose(s1.get_amount("Na+", "count").magnitude, 6.02214e23)



"""
pyEQL salt matching library.

This file contains functions that allow a pyEQL Solution object composed of
individual species (usually ions) to be mapped to a solution of one or more
salts. This mapping is necessary because some parameters (such as activity
coefficient data) can only be determined for salts (e.g. NaCl) and not individual
species (e.g. Na+)

:copyright: 2013-2024 by Ryan S. Kingsbury
:license: LGPL, see LICENSE for more details.

"""

from monty.json import MSONable
from pymatgen.core.ion import Ion

from pyEQL.utils import standardize_formula


class Salt(MSONable):
    """Class to represent a salt."""

    def __init__(self, cation, anion) -> None:
        """
        Create a Salt object based on its component ions.

        Parameters:
            cation, anion: (str) Chemical formula of the cation and anion, respectively.

        Returns:
            Salt : An object representing the properties of the salt

        Examples:
            >>> Salt('Na+','Cl-').formula
            'NaCl'

            >>> Salt('Mg++','Cl-').formula
            'MgCl2'
        """
        # create pymatgen Ion objects
        pmg_cat = Ion.from_formula(cation)
        pmg_an = Ion.from_formula(anion)
        # standardize the cation and anion formulas
        self.cation = standardize_formula(cation)
        self.anion = standardize_formula(anion)

        # get the charges on cation and anion
        self.z_cation = pmg_cat.charge
        self.z_anion = pmg_an.charge

        # assign stoichiometric coefficients by finding a common multiple
        self.nu_cation = int(abs(self.z_anion))
        self.nu_anion = int(abs(self.z_cation))

        # if both coefficients are the same, set each to one
        if self.nu_cation == self.nu_anion:
            self.nu_cation = 1
            self.nu_anion = 1

        # start building the formula, cation first
        salt_formula = ""
        if self.nu_cation > 1:
            # add parentheses if the cation is a polyatomic ion
            if len(pmg_cat.elements) > 1:
                salt_formula += "("
                salt_formula += self.cation.split("[")[0]
                salt_formula += ")"
            else:
                salt_formula += self.cation.split("[")[0]
            salt_formula += str(self.nu_cation)
        else:
            salt_formula += self.cation.split("[")[0]

        if self.nu_anion > 1:
            # add parentheses if the anion is a polyatomic ion
            if len(pmg_an.elements) > 1:
                salt_formula += "("
                salt_formula += self.anion.split("[")[0]
                salt_formula += ")"
            else:
                salt_formula += self.anion.split("[")[0]
            salt_formula += str(self.nu_anion)
        else:
            salt_formula += self.anion.split("[")[0]

        self.formula = salt_formula

    # TODO - consider whether this should be adjusted to be based on total concentrations or not
    # NOTE: speciating the solution results in a decrease in the overall ionic strength, because some of the
    # Mg+2 is converted to monovalent complexes like MgOH+. Hence, the activity coefficients deviate a bit from
    # the published values.
    def get_effective_molality(self, ionic_strength):
        r"""Calculate the effective molality according to [mistry]_.

        .. math:: 2 I \over (\nu_+ z_+^2 + \nu_- z_- ^2)

        Args:
            ionic_strength: Quantity
                The ionic strength of the parent solution, mol/kg

        Returns:
            Quantity: the effective molality of the salt in the parent solution

        References:
            .. [mistry] Mistry, K. H.; Hunter, H. a.; Lienhard V, J. H. Effect of composition and nonideal solution behavior
                on desalination calculations for mixed electrolyte solutions with comparison to seawater. Desalination
                2013, 318, 34-47.
        """
        m_effective = 2 * ionic_strength / (self.nu_cation * self.z_cation**2 + self.nu_anion * self.z_anion**2)

        return m_effective.to("mol/kg")



"""
pyEQL density test suite
============================================

This file contains tests that check the density of electrolyte solutions
that are computed using the Pitzer model to get partial molar volumes.

NOTE: generally, these tests check the module output against experimental
data rather than the theoretical result of the respective functions. In some
cases, the output is also tested against a well-established model published
by USGS(PHREEQC)
"""

import numpy as np
import pyEQL


class Test_density_nacl:
    """
    test Pitzer model for density of NaCl
    ------------------------------------------------
    """

    def test_density_pitzer_nacl_1(self):
        """
        calculate the density at each concentration and compare
        to experimental data

        Published density values at 25 degC over a range of NaCl concentrations up
        to saturation are found in J. Phys. Chem. Reference Data Vol 13 (1), 1984, p.84

        """
        # list of concentrations to test, mol/kg
        conc_list = [0.1, 0.25, 0.5, 0.75, 1, 2, 3, 4, 5, 6]

        # list of published experimental densities
        pub_density = [
            1.00117,
            1.00722,
            1.01710,
            1.02676,
            1.03623,
            1.07228,
            1.10577,
            1.13705,
            1.16644,
            1.1942,
        ]

        for i, conc in enumerate(conc_list):
            conc = str(conc) + "mol/kg"
            sol = pyEQL.Solution()
            sol.add_solute("Na+", conc)
            sol.add_solute("Cl-", conc)
            result = sol.density.to("g/mL").magnitude
            expected = pub_density[i]

            assert np.isclose(result, expected, rtol=0.01)

    def test_density_pitzer_nacl_phreeqc_1(self):
        """
        calculate the density at each concentration and compare
        to the output of the PHREEQC model

        PHREEQC version 3.1.4 was used to calculate density, conductivity, water
        activity, and NaCl activity coefficient for NaCl solutions up to 6m.
        The Pitzer model (pitzer.dat) database was used.
        <http://wwwbrr.cr.usgs.gov/projects/GWC_coupled/phreeqc/>

        """
        # list of concentrations to test, mol/kg
        conc_list = [0.1, 0.25, 0.5, 0.75, 1, 2, 3, 4, 5, 6]

        # list of modeled densities
        phreeqc_pitzer_density = [
            1.00115,
            1.00718,
            1.01702,
            1.02664,
            1.03606,
            1.07204,
            1.10562,
            1.13705,
            1.16651,
            1.19417,
        ]

        for i, conc in enumerate(conc_list):
            conc = str(conc) + "mol/kg"
            sol = pyEQL.Solution()
            sol.add_solute("Na+", conc)
            sol.add_solute("Cl-", conc)
            result = sol.density.to("g/mL").magnitude
            expected = phreeqc_pitzer_density[i]

            assert np.isclose(result, expected, rtol=0.01)



"""
pyEQL volume and concentration methods test suite
=================================================

This file contains tests for the volume and concentration-related methods
used by pyEQL's Solution class
"""

import logging
import platform

import numpy as np
import pytest

from pyEQL import Solution
from pyEQL.engines import PhreeqcEOS

if platform.machine().startswith("arm64") and platform.system().startswith("Darwin"):
    pytest.skip("skipping PHREEQC tests because arm64 is not supported", allow_module_level=True)


@pytest.fixture()
def s1():
    return Solution(volume="2 L", engine="phreeqc")


@pytest.fixture()
def s2():
    return Solution([["Na+", "4 mol/L"], ["Cl-", "4 mol/L"]], volume="2 L", engine="phreeqc")


@pytest.fixture()
def s3():
    return Solution([["Na+", "4 mol/kg"], ["Cl-", "4 mol/kg"]], volume="2 L", engine="phreeqc")


@pytest.fixture()
def s5():
    # 100 mg/L as CaCO3 ~ 1 mM
    return Solution([["Ca+2", "40.078 mg/L"], ["CO3-2", "60.0089 mg/L"]], volume="1 L", engine="phreeqc")


@pytest.fixture()
def s5_pH():
    # 100 mg/L as CaCO3 ~ 1 mM
    return Solution(
        [["Ca+2", "40.078 mg/L"], ["CO3-2", "60.0089 mg/L"]], volume="1 L", balance_charge="pH", engine="phreeqc"
    )


@pytest.fixture()
def s6():
    # non-electroneutral solution with lots of hardness
    # alk = -118 meq/L * 50 = -5900 mg/L, hardness = 12*50 = 600 mg/L as CaCO3
    # charge balance = 2+10+10+10-120-20-12 = -120 meq/L
    return Solution(
        [
            ["Ca+2", "1 mM"],  # 2 meq/L
            ["Mg+2", "5 mM"],  # 10 meq/L
            ["Na+1", "10 mM"],  # 10 meq/L
            ["Ag+1", "10 mM"],  # no contribution to alk or hardness
            ["CO3-2", "6 mM"],  # no contribution to alk or hardness
            ["SO4-2", "60 mM"],  # -120 meq/L
            ["Br-", "20 mM"],
        ],  # -20 meq/L
        volume="1 L",
        engine="phreeqc",
    )


@pytest.fixture()
def s6_Ca():
    # non-electroneutral solution with lots of hardness
    # alk = -118 meq/L * 50 = -5900 mg/L, hardness = 12*50 = 600 mg/L as CaCO3
    # charge balance = 2+10+10+10-120-20-12 = -120 meq/L
    return Solution(
        [
            ["Ca+2", "1 mM"],  # 2 meq/L
            ["Mg+2", "5 mM"],  # 10 meq/L
            ["Na+1", "10 mM"],  # 10 meq/L
            ["Ag+1", "10 mM"],  # no contribution to alk or hardness
            ["CO3-2", "6 mM"],  # no contribution to alk or hardness
            ["SO4-2", "60 mM"],  # -120 meq/L
            ["Br-", "20 mM"],
        ],  # -20 meq/L
        volume="1 L",
        balance_charge="Ca+2",
        engine="phreeqc",
    )


def test_empty_solution_3():
    # create an empty solution
    s1 = Solution(database=None, engine="phreeqc")
    # It should return type Solution
    assert isinstance(s1, Solution)
    # It should have exactly 1L volume
    assert s1.volume.to("L").magnitude == 1.0
    #  the solvent should be water
    assert s1.solvent == "H2O(aq)"
    # It should have 0.997 kg water mass
    assert np.isclose(s1.solvent_mass.to("kg").magnitude, 0.997, atol=1e-3)
    # the temperature should be 25 degC
    assert s1.temperature.to("degC").magnitude == 25
    # the pressure should be 1 atm
    assert s1.pressure.to("atm").magnitude == 1
    # the pH should be 7.0
    assert np.isclose(s1.get_activity("H+"), 1e-7, atol=1e-9)
    # assert np.isclose(s1.pH, 7.0, atol=0.01)
    assert np.isclose(s1.pE, 8.5)
    # it should contain H2O, H+, and OH- species
    assert set(s1.components.keys()) == {"H2O(aq)", "OH[-1]", "H[+1]"}


def test_init_engines():
    """
    Test passing an EOS instance as well as the ideal and native EOS
    """
    s = Solution([["Na+", "4 mol/L"], ["Cl-", "4 mol/L"]], engine="phreeqc")
    assert isinstance(s.engine, PhreeqcEOS)
    assert s.get_activity_coefficient("Na+").magnitude * s.get_activity_coefficient("Cl-").magnitude < 1
    assert s.get_osmotic_coefficient().magnitude == 1
    # with pytest.warns(match="Solute Mg+2 not found"):
    assert s.get_activity_coefficient("Mg+2").magnitude == 1
    assert s.get_activity("Mg+2").magnitude == 0
    s.engine._destroy_ppsol()
    assert s.engine.ppsol is None


def test_conductivity(s1):
    # even an empty solution should have some conductivity
    assert s1.conductivity > 0

    for conc, cond in zip([0.001, 0.05, 0.1], [123.68, 111.01, 106.69]):
        s1 = Solution({"Na+": f"{conc} mol/L", "Cl-": f"{conc} mol/L"})
        assert np.isclose(
            s1.conductivity.to("S/m").magnitude, conc * cond / 10, atol=0.5
        ), f"Conductivity test failed for NaCl at {conc} mol/L. Result = {s1.conductivity.to('S/m').magnitude}"

    # higher concentration data points from Appelo, 2017 Figure 4.
    s1 = Solution({"Na+": "2 mol/kg", "Cl-": "2 mol/kg"})
    assert np.isclose(s1.conductivity.to("mS/cm").magnitude, 145, atol=10)

    # MgCl2
    for conc, cond in zip([0.001, 0.05, 0.1], [124.15, 114.49, 97.05]):
        s1 = Solution({"Mg+2": f"{conc} mol/L", "Cl-": f"{2*conc} mol/L"})
        assert np.isclose(
            s1.conductivity.to("S/m").magnitude, 2 * conc * cond / 10, atol=1
        ), f"Conductivity test failed for MgCl2 at {conc} mol/L. Result = {s1.conductivity.to('S/m').magnitude}"

    # per CRC handbook "standard KCl solutions for calibrating conductiVity cells",
    # 0.1m KCl has a conductivity of 12.824 mS/cm at 25 C
    s_kcl = Solution({"K+": "0.1 mol/kg", "Cl-": "0.1 mol/kg"})
    assert np.isclose(s_kcl.conductivity.magnitude, 1.2824, atol=0.02)  # conductivity is in S/m

    s_kcl.temperature = "5 degC"
    assert np.isclose(s_kcl.conductivity.magnitude, 0.81837, atol=0.06)

    s_kcl.temperature = "50 degC"
    assert np.isclose(s_kcl.conductivity.magnitude, 1.91809, atol=0.18)

    # TODO - conductivity model not very accurate at high conc.
    s_kcl = Solution({"K+": "1 mol/kg", "Cl-": "1 mol/kg"})
    assert np.isclose(s_kcl.conductivity.magnitude, 10.862, rtol=0.05)


def test_equilibrate(s1, s2, s5_pH, s6_Ca, caplog):
    assert "H2(aq)" not in s1.components
    orig_pH = s1.pH
    orig_pE = s1.pE
    orig_mass = s1.mass
    orig_density = s2.density.magnitude
    orig_solv_mass = s2.solvent_mass.magnitude
    s1.equilibrate()
    assert "H2(aq)" in s1.components
    assert np.isclose(s1.charge_balance, 0, atol=1e-8)
    assert np.isclose(s1.pH, orig_pH, atol=0.01)
    assert np.isclose(s1.pE, orig_pE)

    assert "NaOH(aq)" not in s2.components
    orig_density = s2.density.magnitude
    orig_solv_mass = s2.solvent_mass.magnitude
    orig_pH = s2.pH
    orig_pE = s2.pE
    orig_mass = s2.mass
    s2.equilibrate()
    assert np.isclose(s2.mass, orig_mass)
    assert np.isclose(s2.density.magnitude, orig_density)
    assert np.isclose(s2.solvent_mass.magnitude, orig_solv_mass)
    assert "NaOH(aq)" in s2.components

    # total element concentrations should be conserved after equilibrating
    assert np.isclose(s2.get_total_amount("Na", "mol").magnitude, 8)
    assert np.isclose(s2.get_total_amount("Cl", "mol").magnitude, 8)
    assert np.isclose(s2.solvent_mass.magnitude, orig_solv_mass)
    assert np.isclose(s2.density.magnitude, orig_density)
    # the pH should drop due to hydrolysis of Ca+2
    assert s2.pH < orig_pH
    assert np.isclose(s2.pE, orig_pE)
    assert np.isclose(s2.mass, orig_mass)
    # this solution has balance_charge=None, therefore, the charge balance
    # may be off after equilibration
    assert not np.isclose(s2.charge_balance, 0, atol=1e-8)
    eq_Hplus = s2.components["H+"]
    s2.balance_charge = "pH"
    s2.equilibrate()
    assert np.isclose(s2.charge_balance, 0, atol=1e-8)
    assert s2.components["H+"] > eq_Hplus

    # test log message if there is a species not present in the phreeqc database
    s_zr = Solution(
        {"Zr+4": "0.05 mol/kg", "Na+": "0.05 mol/kg", "Cl-": "0.1 mol/kg"}, engine="phreeqc", log_level="WARNING"
    )
    totzr = s_zr.get_total_amount("Zr", "mol")
    with caplog.at_level(logging.WARNING, "pyEQL.engines"):
        s_zr.equilibrate()
        assert "likely absent from its database" in caplog.text
        assert "Zr[+4]" in s_zr.components
        assert s_zr.get_total_amount("Zr", "mol") == totzr

    # this solution is the only one in the test that contains alkalinity
    # and equilibrating it results in a shift in the pH
    # the CO3-2 initially present reacts with the water to consume H+ and
    # increase the pH by approximately 0.0006 M (b/c at pH 7 virtually all
    # carbonate is present as HCO3-) -log10(0.001) =
    assert "HCO3[-1]" not in s5_pH.components
    assert np.isclose(s5_pH.charge_balance, 0)
    orig_pH = s5_pH.pH
    orig_pE = s5_pH.pE
    orig_density = s5_pH.density.magnitude
    orig_solv_mass = s5_pH.solvent_mass.magnitude
    set(s5_pH.components.keys())
    s5_pH.equilibrate()
    assert np.isclose(s5_pH.get_total_amount("Ca", "mol").magnitude, 0.001)
    assert np.isclose(s5_pH.get_total_amount("C(4)", "mol").magnitude, 0.001)
    # due to the large pH shift, water mass and density need not be perfectly conserved
    assert np.isclose(s5_pH.solvent_mass.magnitude, orig_solv_mass, atol=1e-3)
    assert np.isclose(s5_pH.density.magnitude, orig_density, atol=1e-3)
    assert np.isclose(s5_pH.charge_balance, 0)
    assert "CaOH[+1]" in s5_pH.components
    assert "HCO3[-1]" in s5_pH.components
    s5_pH_after = s5_pH.pH
    assert s5_pH_after > orig_pH
    assert np.isclose(s5_pH.pE, orig_pE)

    # repeated calls to equilibrate should not change the properties (much)
    for i in range(10):
        s5_pH.equilibrate()
        assert np.isclose(s5_pH.charge_balance, 0, atol=1e-8), f"C.B. failed at iteration {i}"
        assert np.isclose(s5_pH.pH, s5_pH_after, atol=0.01), f"pH failed at iteration {i}"
        assert np.isclose(s5_pH.pE, orig_pE), f"pE failed at iteration {i}"

    # test equilibrate() with a non-pH balancing species
    assert np.isclose(s6_Ca.charge_balance, 0, atol=1e-8)
    initial_Ca = s6_Ca.get_total_amount("Ca", "mol").magnitude
    assert s6_Ca.balance_charge == "Ca[+2]"
    s6_Ca.equilibrate()
    assert s6_Ca.get_total_amount("Ca", "mol").magnitude != initial_Ca
    assert np.isclose(s6_Ca.charge_balance, 0, atol=1e-8)



"""
pyEQL methods for chemical equilibrium calculations (e.g. acid/base, reactions,
redox, complexation, etc.).

NOTE: these methods are not currently used but are here for the future.

:copyright: 2013-2024 by Ryan S. Kingsbury
:license: LGPL, see LICENSE for more details.

"""

# import libraries for scientific functions
import logging

import numpy as np

from pyEQL import ureg

logger = logging.getLogger(__name__)

# TODO - not used. Remove?
SPECIES_ALIAISES = {
    "Sodium": "Na+",
    "Potassium": "K+",
    "Calcium": "Ca+2",
    "Barium": "Ba+2",
    "Strontium": "Sr+2",
    "Magnesium": "Mg+2",
    "Chloride": "Cl-",
    "Fluoride": "F-",
    "Nitrate": "NO3-",
    "Ammonium": "NH4+",
    "Sulfate": "SO4-2",
    "Phosphate": "PO4-3",
    "Carbonate": "CO3-2",
    "Bicarbonate": "HCO3-",
}


def adjust_temp_pitzer(c1, c2, c3, c4, c5, temp, temp_ref=ureg.Quantity("298.15 K")):
    """
    Calculate a parameter for the Pitzer model based on temperature-dependent coefficients c1,c2,c3,c4,and c5.

    Args:
        c1, c2, c3, c4, c5 (float): Temperature-dependent coefficients for the pitzer parameter of interest.
        temp (Quantity): The temperature at which the Pitzer parameter is to be calculated.
        temp_ref (Quantity, optional): The reference temperature on which the parameters are based. Defaults to 298.15 K if omitted.

    Note:
        As described in the PHREEQC documentation.
    """
    return (
        c1
        + c2 * (1 / temp + 1 / temp_ref)
        + c2 * np.log(temp / temp_ref)
        + c3 * (temp - temp_ref)
        + c4 * (temp**2 - temp_ref**2)
        + c5 * (temp**-2 - temp_ref**-2)
    )


def adjust_temp_vanthoff(equilibrium_constant, enthalpy, temperature, reference_temperature=ureg.Quantity(25, "degC")):
    r"""
    Adjust a reaction equilibrium constant from one temperature to another.

    Args:
        equilibrium_constant (float): The reaction equilibrium constant for the reaction.
        enthalpy (Quantity): The enthalpy change (delta H) for the reaction in kJ/mol. Assumed independent of temperature (see Notes).
        temperature (Quantity): The desired reaction temperature in degrees Celsius.
        reference_temperature (Quantity, optional): The temperature at which equilibrium_constant is valid. Defaults to 25 degrees C if omitted.

    Returns:
        float: The adjusted reaction equilibrium constant.

    Note:
        This function implements the Van't Hoff equation to adjust measured equilibrium constants to other temperatures.

        .. math::
            ln(K2 / K1) = {\delta H \over R} ( {1 \over T_1} - {1 \over T_2} )

        This implementation assumes that the enthalpy is independent of temperature over the range of interest.

    References:
        Stumm, Werner and Morgan, James J. Aquatic Chemistry, 3rd ed, pp 53. Wiley Interscience, 1996.

    Examples:
        >>> adjust_temp_vanthoff(0.15,ureg.Quantity('-197.6 kJ/mol'),ureg.Quantity('42 degC'),ureg.Quantity(' 25degC')) #doctest: +ELLIPSIS
        0.00203566...

        If the 'ref_temperature' parameter is omitted, a default of 25 C is used.

        >>> adjust_temp_vanthoff(0.15,ureg.Quantity('-197.6 kJ/mol'),ureg.Quantity('42 degC')) #doctest: +ELLIPSIS
        0.00203566...
    """
    output = equilibrium_constant * np.exp(
        enthalpy / ureg.R * (1 / reference_temperature.to("K") - 1 / temperature.to("K"))
    )

    logger.debug(
        "Adjusted equilibrium constant K=%s from %s to %s degrees Celsius with Delta H = %s. Adjusted K = %s % equilibrium_constant,reference_temperature,temperature,enthalpy,output"
    )

    logger.info(
        "Note that the Van't Hoff equation assumes enthalpy is independent of temperature over the range of interest"
    )
    return output


def adjust_temp_arrhenius(
    rate_constant,
    activation_energy,
    temperature,
    reference_temperature=ureg.Quantity(25, "degC"),
):
    r"""
    Adjust a reaction equilibrium constant from one temperature to another.

    Args:
        rate_constant (Quantity): The parameter value (usually a rate constant) being adjusted.
        activation_energy (Quantity): The activation energy of the process, in kJ/mol.
        temperature (Quantity): The desired reaction temperature.
        reference_temperature (Quantity, optional): The temperature at which equilibrium_constant is valid. Defaults to 25 degrees C if omitted.

    Returns:
        Quantity: The adjusted reaction equilibrium constant.

    Notes:
        This function implements the Arrhenius equation to adjust measured rate constants to other temperatures.
        TODO - add better reference

        .. math::
            ln(\frac{K2}{K1}) = \frac{E_a}{R} ( \frac{1}{T_{1}} - {\frac{1}{T_2}} )

    References:
        http://chemwiki.ucdavis.edu/Physical_Chemistry/Kinetics/Reaction_Rates/Temperature_Dependence_of_Reaction_Rates/Arrhenius_Equation

    Examples:
        >>> adjust_temp_arrhenius(7,900*ureg.Quantity('kJ/mol'),37*ureg.Quantity('degC'),97*ureg.Quantity('degC')) #doctest: +ELLIPSIS
        1.8867225...e-24
    """
    output = rate_constant * np.exp(
        activation_energy / ureg.R * (1 / reference_temperature.to("K") - 1 / temperature.to("K"))
    )

    logger.debug(
        f"Adjusted parameter {rate_constant} from {reference_temperature} to {temperature} degrees Celsius with"
        f"Activation Energy = {activation_energy}s kJ/mol. Adjusted value = {output}"
    )

    return output


def alpha(n, pH, pKa_list):
    r"""
    Returns the acid-base distribution coefficient (alpha) of an acid in the n-deprotonated form at a given pH.

    Args:
        n (int): The number of protons that have been lost by the desired form of the acid. Also the subscript on the
            alpha value. E.g. for bicarbonate (HCO3-), n=1 because 1 proton has been lost from the fully-protonated
            carbonic acid (H2CO3) form.
        pH (float or int): The pH of the solution.
        pKa_list (list of floats or ints): The pKa values (negative log of equilibrium constants) for the acid of
            interest. There must be a minimum of n pKa values in the list.

    Returns:
        float: The fraction of total acid present in the specified form.

    Notes:
        The acid-base coefficient is calculated as follows: [stm]_

        .. math::

            \alpha_n = \frac{term_n}{[H+]^n + k_{a1}[H+]^{n-1} + k_{a1}k_{a2}[H+]^{n-2} ... k_{a1}k_{a2}...k_{an} }

        Where :math: '\term_n' refers to the nth term in the denominator, starting from 0

    References:
        Stumm, Werner and Morgan, James J. Aquatic Chemistry, 3rd ed, pp 127-130. Wiley Interscience, 1996.

    Examples:
        >>> alpha(1,8,[4.7]) #doctest: +ELLIPSIS
        0.999...

        The sum of all alpha values should equal 1

        >>> alpha(0,8,[6.35,10.33]) #doctest: +ELLIPSIS
        0.021...
        >>> alpha(1,8,[6.35,10.33]) #doctest: +ELLIPSIS
        0.979...
        >>> alpha(2,8,[6.35,10.33]) #doctest: +ELLIPSIS
        2.043...e-09

        If pH is equal to one of the pKa values the function should return 0.5.

        >>> alpha(1,6.35,[6.35,10.33])
        0.5
    """
    # generate an error if no pKa values are specified
    if len(pKa_list) == 0:
        raise ValueError("No pKa values given. Cannot calculate distribution coefficient.")

    # generate an error if n > number of pKa values
    if len(pKa_list) < n:
        raise ValueError("Insufficient number of pKa values given. Cannot calculate distribution coefficient.")

    # sort the pKa list
    pKa_list = sorted(pKa_list)

    # determine how many protons the acid has
    num_protons = len(pKa_list)

    # build a list of terms where the term subscript corresponds to the list index
    terms_list = [10 ** (-pH * num_protons)]
    for i in range(len(pKa_list)):
        # multiply together all K values with
        k_term = 1
        for pK in pKa_list[0 : i + 1]:
            k_term *= 10**-pK
        terms_list.append(k_term * 10 ** (-pH * (num_protons - (i + 1))))
    print(terms_list)

    # return the desired distribution factor
    alpha = terms_list[n] / sum(terms_list)
    logger.debug(
        "Calculated %s-deprotonated acid distribution coefficient of %s for pKa=%s at pH %s % n,alpha,pKa_list,pH"
    )
    return alpha



"""
pyEQL utilities

:copyright: 2013-2024 by Ryan S. Kingsbury
:license: LGPL, see LICENSE for more details.

"""

import logging
from collections import UserDict
from functools import lru_cache
from typing import Any

from iapws import IAPWS95, IAPWS97
from pymatgen.core.ion import Ion

from pyEQL import ureg

logger = logging.getLogger(__name__)


def interpret_units(unit: str) -> str:
    """
    Translate commonly used environmental units such as 'ppm' into strings that `pint` can understand.

    Args:
        unit: string representing the unit to translate

    Returns: a unit that pint can understand
    """
    if unit == "m":  # molal
        return "mol/kg"
    if unit == "ppm":
        return "mg/L"
    if unit == "ppb":
        return "ug/L"
    if unit == "ppt":
        return "ng/L"
    # if all else fails, return the unit we were provided
    return unit


@lru_cache
def standardize_formula(formula: str):
    """
    Convert a chemical formula into standard form.

    Args:
        formula: the chemical formula to standardize.

    Returns:
        A standardized chemical formula

    Raises:
        ValueError if `formula` cannot be processed or is invalid.

    Notes:
        Currently this method standardizes formulae by passing them through `pymatgen.core.ion.Ion.reduced_formula()`.
        For ions, this means that 1) the charge number will always be listed explicitly and 2) the charge number will
        be enclosed in square brackets to remove any ambiguity in the meaning of the formula. For example, 'Na+',
        'Na+1', and 'Na[+]' will all standardize to "Na[+1]"
    """
    # fix permuted sign and charge number (e.g. Co2+)
    for str, rep in zip(["²⁺", "³⁺", "⁴⁺", "²⁻", "³⁻", "⁴⁻"], ["+2", "+3", "+4", "-2", "-3", "-4"]):
        formula = formula.replace(str, rep)

    # replace superscripts with non superscripts
    for char, rep in zip("⁻⁺⁰¹²³⁴⁵⁶⁷⁸⁹", "-+0123456789"):
        formula = formula.replace(char, rep)

    # replace subscripts with non subscripts
    for char, rep in zip("₀₁₂₃₄₅₆₇₈₉", "0123456789"):
        formula = formula.replace(char, rep)

    sform = Ion.from_formula(formula).reduced_formula

    # TODO - manual formula adjustments. May be implemented upstream in pymatgen in the future
    # thanks to @xiaoxiaozhu123 for pointing out these issues in
    # https://github.com/KingsburyLab/pyEQL/issues/136

    # ammonia
    if sform == "H4N[+1]":
        sform = "NH4[+1]"
    elif sform == "H3N(aq)":
        sform = "NH3(aq)"
    # phosphoric acid system
    elif sform == "PH3O4(aq)":
        sform = "H3PO4(aq)"
    elif sform == "PHO4[-2]":
        sform = "HPO4[-2]"
    elif sform == "P(HO2)2[-1]":
        sform = "H2PO4[-1]"
    # thiocyanate
    elif sform == "CSN[-1]":
        sform = "SCN[-1]"
    # triiodide, nitride, an phosphide
    elif sform == "I[-0.33333333]":
        sform = "I3[-1]"
    elif sform == "N[-0.33333333]":
        sform = "N3[-1]"
    elif sform == "P[-0.33333333]":
        sform = "P3[-1]"
    # formate
    elif sform == "HCOO[-1]":
        sform = "HCO2[-1]"
    # oxalate
    elif sform == "CO2[-1]":
        sform = "C2O4[-2]"
    # triflate
    elif sform == "CS(OF)3[-1]":
        sform = "CF3SO3[-1]"
    # haloacetic acids of F, Cl, Br, I
    elif sform == "C2Cl3O2[-1]":
        sform = "CCl3COO[-1]"
    elif sform == "C2O2F3[-1]":
        sform = "CF3COO[-1]"
    elif sform == "C2I3O2[-1]":
        sform = "CI3COO[-1]"
    elif sform == "C2Br3O2[-1]":
        sform = "CBr3COO[-1]"

    # Cl+F
    elif sform == "C2Cl2O2F[-1]":
        sform = "CFCl2COO[-1]"
    elif sform == "C2Cl(OF)2[-1]":
        sform = "CF2ClCOO[-1]"

    # Cl+Br
    elif sform == "C2Br(ClO)2[-1]":
        sform = "CBrCl2COO[-1]"
    elif sform == "C2Br2ClO2[-1]":
        sform = "CBr2ClCOO[-1]"

    # Cl+I
    elif sform == "C2I(ClO)2[-1]":
        sform = "CICl2COO[-1]"
    elif sform == "C2I2ClO2[-1]":
        sform = "CI2ClCOO[-1]"

    # TODO - consider adding recognition of special formulas like MeOH for methanol or Cit for citrate
    return sform


def format_solutes_dict(solute_dict: dict, units: str):
    """
    Formats a dictionary of solutes by converting the amount to a string with the provided units suitable for passing to
    use with the Solution class. Note that all solutes must be given in the same units.

    Args:
        solute_dict: The dictionary to format. This must be of the form dict{str: Number}
            e.g. {"Na+": 0.5, "Cl-": 0.9}
        units: The units to use for the solute. e.g. "mol/kg"

    Returns:
        A formatted solute dictionary.

    Raises:
        TypeError if `solute_dict` is not a dictionary.
    """
    if not isinstance(solute_dict, dict):
        raise TypeError("solute_dict must be a dictionary. Refer to the doc for proper formatting.")

    return {key: f"{value!s} {units}" for key, value in solute_dict.items()}


@lru_cache
@ureg.wraps(ret=None, args=["K", "MPa"], strict=False)
def create_water_substance(temperature: float, pressure: float):
    """
    Instantiate a water substance model from IAPWS.

    Args:
        temperature: the desired temperature in K
        pressure: the desired pressure in MPa

    Notes:
        The IAPWS97 model is much faster than IAPWS95, but the latter can do temp
        below zero. See https://github.com/jjgomera/iapws/issues/14. Hence,
        IAPWS97 will be used except when `temperature` is less than 0 degC.

    Returns:
        A IAPWS97 or IAPWS95 instance
    """
    if temperature >= 273.15:
        return IAPWS97(T=temperature, P=pressure)
    return IAPWS95(T=temperature, P=pressure)


class FormulaDict(UserDict):
    """
    Automatically converts keys on get/set using pymatgen.core.Ion.from_formula(key).reduced_formula.

    This allows getting/setting/updating of Solution.components using flexible
    formula notation (e.g., "Na+", "Na+1", "Na[+]" all have the same effect)
    """

    def __getitem__(self, key) -> Any:
        return super().__getitem__(standardize_formula(key))

    def __setitem__(self, key, value) -> None:
        super().__setitem__(standardize_formula(key), value)
        # sort contents anytime an item is set
        self.data = dict(sorted(self.items(), key=lambda x: x[1], reverse=True))

    # Necessary to define this so that .get() works properly in python 3.12+
    # see https://github.com/python/cpython/issues/105524
    def __contains__(self, key) -> bool:
        return standardize_formula(key) in self.data

    def __delitem__(self, key) -> None:
        super().__delitem__(standardize_formula(key))



"""
pyEQL activity correction methods test suite
============================================

This file contains tests for some of the activity correction methods
employed by pyEQL.

NOTE: generally, these tests check the module output against experimental
data rather than the theoretical result of the respective functions. In some
cases, the output is also tested against a well-established model published
by USGS(PHREEQC)
"""
import platform

import numpy as np
import pytest

from pyEQL.activity_correction import _debye_parameter_activity, _debye_parameter_B
from pyEQL.solution import Solution

## Tests of the pitzer model


def test_units_and_equality():
    s1 = Solution([["Na+", "0.1 mol/L"], ["Cl-", "0.1 mol/L"]])

    # TODO - add a quantitative test for molal, molar, and rational conversion
    # the activity coefficient should be dimensionless
    assert s1.get_activity_coefficient("Na+").dimensionality == ""
    assert s1.get_activity_coefficient("Na+", scale="molar").dimensionality == ""
    assert s1.get_activity_coefficient("Na+", scale="rational").dimensionality == ""

    assert s1.get_activity("Na+").dimensionality == ""
    assert s1.get_activity("Na+", scale="molar").dimensionality == ""
    assert s1.get_activity("Na+", scale="rational").dimensionality == ""

    assert s1.get_osmotic_coefficient(scale="molal").dimensionality == ""
    assert s1.get_osmotic_coefficient(scale="rational").dimensionality == ""
    assert s1.get_osmotic_coefficient(scale="fugacity").dimensionality == ""
    with pytest.raises(ValueError, match="Invalid scale argument"):
        s1.get_activity_coefficient("Na+", scale="random")
    with pytest.raises(ValueError, match="Invalid scale argument"):
        s1.get_activity("Na+", scale="random")
    with pytest.raises(ValueError, match="Invalid scale argument"):
        s1.get_osmotic_coefficient(scale="random")

    # the activity should be dimensionless
    assert s1.get_activity("Na+").dimensionality == ""
    assert s1.get_activity("H2O").dimensionality == ""

    # the activity coefficient of both the Na+ and Cl- should be the same
    a1 = s1.get_activity_coefficient("Na+")
    a2 = s1.get_activity_coefficient("Cl-")
    assert a1 == a2


def test_debye_params():
    # tests of the various Debye Huckel parameters
    # A should be equal to 0.509 at 25 C, for log base 10
    assert np.isclose(_debye_parameter_activity().magnitude / 2.303, 0.509, atol=1e-3)
    assert np.isclose(_debye_parameter_B().to("nm**-1 * kg**0.5/mol**0.5").magnitude, 3.29, atol=1e-2)


def test_activity_crc_HCl():
    """
    calculate the activity coefficient of HCl at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *CRC Handbook of Chemistry and Physics*, Mean Activity Coefficients of Electrolytes as a Function of Concentration,
    in: W.M. Haynes (Ed.), 92nd ed., 2011.

    """
    cation = "H+"
    nu_cation = 1
    anion = "Cl-"
    nu_anion = 1

    # list of concentrations to test, mol/kg
    conc_list = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.965,
        0.952,
        0.929,
        0.905,
        0.876,
        0.832,
        0.797,
        0.768,
        0.759,
        0.811,
        1.009,
        2.380,
    ]

    for i, conc in enumerate(conc_list):
        conc_c = str(conc * nu_cation) + "mol/kg"
        conc_a = str(conc * nu_anion) + "mol/kg"
        sol = Solution()
        sol.add_solute(cation, conc_c)
        sol.add_solute(anion, conc_a)
        act_cat = sol.get_activity_coefficient(cation)
        act_an = sol.get_activity_coefficient(anion)
        result = (act_cat**nu_cation * act_an**nu_anion) ** (1 / (nu_cation + nu_anion))
        expected = pub_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


def test_activity_crc_CsI():
    """
    calculate the activity coefficient of CsI at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *CRC Handbook of Chemistry and Physics*, Mean Activity Coefficients of Electrolytes as a Function of Concentration,
    in: W.M. Haynes (Ed.), 92nd ed., 2011.

    """
    cation = "Cs+"
    nu_cation = 1
    anion = "I-"
    nu_anion = 1

    # list of concentrations to test, mol/kg
    conc_list = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.965,
        0.951,
        0.925,
        0.898,
        0.863,
        0.804,
        0.749,
        0.688,
        0.601,
        0.534,
        0.470,
    ]

    for i, conc in enumerate(conc_list):
        conc_c = str(conc * nu_cation) + "mol/kg"
        conc_a = str(conc * nu_anion) + "mol/kg"
        sol = Solution()
        sol.add_solute(cation, conc_c)
        sol.add_solute(anion, conc_a)
        act_cat = sol.get_activity_coefficient(cation)
        act_an = sol.get_activity_coefficient(anion)
        result = (act_cat**nu_cation * act_an**nu_anion) ** (1 / (nu_cation + nu_anion))
        expected = pub_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


def test_activity_crc_bacl2():
    """
    calculate the activity coefficient of BaCl2 at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *CRC Handbook of Chemistry and Physics*, Mean Activity Coefficients of Electrolytes as a Function of Concentration,
    in: W.M. Haynes (Ed.), 92nd ed., 2011.

    """
    # list of concentrations to test, mol/kg
    conc_list = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.887,
        0.849,
        0.782,
        0.721,
        0.653,
        0.559,
        0.492,
        0.436,
        0.391,
        0.393,
    ]

    for i, conc in enumerate(conc_list):
        conc_c = str(conc) + "mol/kg"
        conc_a = str(conc * 2) + "mol/kg"
        sol = Solution()
        sol.add_solute("Ba+2", conc_c)
        sol.add_solute("Cl-", conc_a)
        act_cat = sol.get_activity_coefficient("Ba+2")
        act_an = sol.get_activity_coefficient("Cl-")
        result = (act_cat**1 * act_an**2) ** (1 / 3)
        expected = pub_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


def test_activity_crc_licl():
    """
    calculate the activity coefficient of LiCl at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *CRC Handbook of Chemistry and Physics*, Mean Activity Coefficients of Electrolytes as a Function of Concentration,
    in: W.M. Haynes (Ed.), 92nd ed., 2011.

    """
    # list of concentrations to test, mol/kg
    conc_list = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.965,
        0.952,
        0.928,
        0.904,
        0.874,
        0.827,
        0.789,
        0.756,
        0.739,
        0.775,
        0.924,
        2.0,
    ]

    for i, conc in enumerate(conc_list):
        conc_c = str(conc) + "mol/kg"
        conc_a = str(conc) + "mol/kg"
        sol = Solution()
        sol.add_solute("Li+", conc_c)
        sol.add_solute("Cl-", conc_a)
        act_cat = sol.get_activity_coefficient("Li+")
        act_an = sol.get_activity_coefficient("Cl-")
        result = (act_cat**1 * act_an**1) ** (1 / 2)
        expected = pub_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


def test_activity_crc_rbcl():
    """
    calculate the activity coefficient of RbCl at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *CRC Handbook of Chemistry and Physics*, Mean Activity Coefficients of Electrolytes as a Function of Concentration,
    in: W.M. Haynes (Ed.), 92nd ed., 2011.

    """
    # list of concentrations to test, mol/kg
    conc_list = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.965,
        0.951,
        0.926,
        0.900,
        0.867,
        0.811,
        0.761,
        0.707,
        0.633,
        0.583,
        0.546,
        0.544,
    ]

    for i, conc in enumerate(conc_list):
        conc_c = str(conc) + "mol/kg"
        conc_a = str(conc) + "mol/kg"
        sol = Solution()
        sol.add_solute("Rb+", conc_c)
        sol.add_solute("Cl-", conc_a)
        result = sol.get_activity_coefficient("Rb+")
        expected = pub_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


@pytest.mark.skipif(platform.machine() == "arm64" and platform.system() == "Darwin", reason="arm64 not supported")
def test_activity_crc_MgCl2():
    """
    calculate the activity coefficient of MgCl2 at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *CRC Handbook of Chemistry and Physics*, Mean Activity Coefficients of Electrolytes as a Function of Concentration,
    in: W.M. Haynes (Ed.), 92nd ed., 2011.

    """
    cation = "Mg+2"
    nu_cation = 1
    anion = "Cl-"
    nu_anion = 2

    # list of concentrations to test, mol/kg
    conc_list = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.889,
        0.852,
        0.790,
        0.734,
        0.672,
        0.590,
        0.535,
        0.493,
        0.485,
        0.577,
        1.065,
        14.40,
    ]

    for i, conc in enumerate(conc_list):
        conc_c = str(conc * nu_cation) + "mol/kg"
        conc_a = str(conc * nu_anion) + "mol/kg"
        sol = Solution()
        sol.add_solute(cation, conc_c)
        sol.add_solute(anion, conc_a)
        act_cat = sol.get_activity_coefficient(cation)
        act_an = sol.get_activity_coefficient(anion)
        result = (act_cat**nu_cation * act_an**nu_anion) ** (1 / (nu_cation + nu_anion))
        expected = pub_activity_coeff[i]
        print(sol.ionic_strength)

        assert np.isclose(result, expected, rtol=0.05)

        # ignore the highest concentration; I=8.35 when speciated vs. 15 without speciation
        if conc <= 2:
            sol.equilibrate()
            # NOTE: speciating the solution results in a decrease in the overall ionic strength, because some of the
            # Mg+2 is converted to monovalent complexes like MgOH+. Hence, the activity coefficients deviate a bit from
            # the published values.
            print(sol.ionic_strength)
            act_cat = sol.get_activity_coefficient(cation)
            act_an = sol.get_activity_coefficient(anion)
            result = (act_cat**nu_cation * act_an**nu_anion) ** (1 / (nu_cation + nu_anion))
            assert np.isclose(result, expected, rtol=0.05 * (1 + (4 * conc)))  # max rtol = 0.45


def test_activity_crc_KBr():
    """
    calculate the activity coefficient of KBr at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *CRC Handbook of Chemistry and Physics*, Mean Activity Coefficients of Electrolytes as a Function of Concentration,
    in: W.M. Haynes (Ed.), 92nd ed., 2011.

    """
    cation = "K+"
    nu_cation = 1
    anion = "Br-"
    nu_anion = 1

    # list of concentrations to test, mol/kg
    conc_list = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1, 2, 5]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.965,
        0.952,
        0.927,
        0.902,
        0.870,
        0.817,
        0.771,
        0.722,
        0.658,
        0.617,
        0.593,
        0.626,
    ]

    for i, conc in enumerate(conc_list):
        conc_c = str(conc * nu_cation) + "mol/kg"
        conc_a = str(conc * nu_anion) + "mol/kg"
        sol = Solution()
        sol.add_solute(cation, conc_c)
        sol.add_solute(anion, conc_a)
        act_cat = sol.get_activity_coefficient(cation)
        act_an = sol.get_activity_coefficient(anion)
        result = (act_cat**nu_cation * act_an**nu_anion) ** (1 / (nu_cation + nu_anion))
        expected = pub_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


def test_activity_crc_k2so4():
    """
    calculate the activity coefficient of K2SO4 at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *CRC Handbook of Chemistry and Physics*, Mean Activity Coefficients of Electrolytes as a Function of Concentration,
    in: W.M. Haynes (Ed.), 92nd ed., 2011.

    """
    # list of concentrations to test, mol/kg
    conc_list = [0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.885,
        0.844,
        0.772,
        0.704,
        0.625,
        0.511,
        0.424,
        0.343,
        0.251,
    ]

    for i, conc in enumerate(conc_list):
        conc_c = str(conc * 2) + "mol/kg"
        conc_a = str(conc) + "mol/kg"
        sol = Solution()
        sol.add_solute("K+", conc_c)
        sol.add_solute("SO4-2", conc_a)
        act_cat = sol.get_activity_coefficient("K+")
        act_an = sol.get_activity_coefficient("SO4-2")
        result = (act_cat**2 * act_an**1) ** (1 / 3)
        expected = pub_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


def test_activity_pitzer_nacl_1():
    """
    calculate the activity coefficient at each concentration and compare
    to experimental data

    Experimental activity coefficient values at 25 degC are found in
    *J. Phys. Chem. Reference Data* Vol 13 (1), 1984, p.53.

    """
    # list of concentrations to test, mol/kg
    conc_list = [0.1, 0.25, 0.5, 0.75, 1, 2, 3, 4, 5, 6]

    # list of published experimental activity coefficients
    pub_activity_coeff = [
        0.778,
        0.72,
        0.681,
        0.665,
        0.657,
        0.669,
        0.714,
        0.782,
        0.873,
        0.987,
    ]

    for i, conc in enumerate(conc_list):
        conc = str(conc) + "mol/kg"
        sol = Solution()
        sol.add_solute("Na+", conc)
        sol.add_solute("Cl-", conc)
        result = sol.get_activity_coefficient("Na+")
        expected = pub_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


# The pitzer model diverges a bit from experimental data at high concentration
@pytest.mark.xfail()
def test_water_activity_pitzer_nacl_1():
    r"""
    calculate the water activity at each concentration and compare
    to experimental data

    Experimental osmotic coefficients for NaCl are found in:
    Pitzer and Pelper, 1984. "Thermodyamic Properties of Aqueous Sodium Chloride Solutions"
    *J. Phys. Chem. Ref. Data* 13(1).

    Osmotic coefficients were converted into water activity according to the equation
    found in
    Blandamer, Mike J., Engberts, Jan B. F. N., Gleeson, Peter T., Reis,
    Joao Carlos R., 2005. "Activity of water in aqueous systems: A frequently
    neglected property." *Chemical Society Review* 34, 440-458.

    .. math:: ln a_w = - \Phi M_w \sum_i m_i

    Where :math:`M_w` is the molar mass of water (0.018015 kg/mol) and :math:`m_i` is the molal concentration
    of each species.

    """
    # list of concentrations to test, mol/kg
    conc_list = [0.1, 0.25, 0.5, 0.75, 1, 2, 3, 4, 5, 6]

    # list of published experimental water activity
    pub_water_activity = [
        0.9964,
        0.9910,
        0.9821,
        0.9733,
        0.9646,
        0.9304,
        0.8975,
        0.8657,
        0.8351,
        0.8055,
    ]

    for i, conc in enumerate(conc_list):
        conc = str(conc) + "mol/kg"
        sol = Solution()
        sol.add_solute("Na+", conc)
        sol.add_solute("Cl-", conc)
        result = sol.get_water_activity()
        expected = pub_water_activity[i]

        assert np.isclose(result, expected, rtol=0.05)


def test_activity_pitzer_phreeqc_nacl_2():
    """
    calculate the activity coefficient at each concentration and compare
    to the output of the PHREEQC model

    PHREEQC version 3.1.4 was used to calculate density, conductivity, water
    activity, and NaCl activity coefficient for NaCl solutions up to 6m.
    The Pitzer model (pitzer.dat) database was used.
    <http://wwwbrr.cr.usgs.gov/projects/GWC_coupled/phreeqc/>

    """
    # list of concentrations to test, mol/kg
    conc_list = [0.1, 0.25, 0.5, 0.75, 1, 2, 3, 4, 5, 6]

    # list of modeled activity coefficients
    phreeqc_pitzer_activity_coeff = [
        0.7771,
        0.7186,
        0.6799,
        0.6629,
        0.6561,
        0.6675,
        0.7128,
        0.7825,
        0.8729,
        0.9874,
    ]

    for i, conc in enumerate(conc_list):
        conc = str(conc) + "mol/kg"
        sol = Solution()
        sol.add_solute("Na+", conc)
        sol.add_solute("Cl-", conc)
        result = sol.get_activity_coefficient("Na+")
        expected = phreeqc_pitzer_activity_coeff[i]

        assert np.isclose(result, expected, rtol=0.05)


def test_water_activity_phreeqc_pitzer_nacl_2():
    """
    calculate the water activity at each concentration and compare
    to the output of the PHREEQC model

    PHREEQC version 3.1.4 was used to calculate density, conductivity, water
    activity, and NaCl activity coefficient for NaCl solutions up to 6m.
    The Pitzer model (pitzer.dat) database was used.
    <http://wwwbrr.cr.usgs.gov/projects/GWC_coupled/phreeqc/>

    """
    # list of concentrations to test, mol/kg
    conc_list = [0.1, 0.25, 0.5, 0.75, 1, 2, 3, 4, 5, 6]

    # list of modeled water activities
    phreeqc_pitzer_water_activity = [
        0.997,
        0.992,
        0.984,
        0.975,
        0.967,
        0.932,
        0.893,
        0.851,
        0.807,
        0.759,
    ]

    for i, conc in enumerate(conc_list):
        sol = Solution(
            {
                "Na+": f"{conc} mol/kg",
                "Cl-": f"{conc} mol/kg",
            }
        )
        result = sol.get_water_activity()
        expected = phreeqc_pitzer_water_activity[i]

        assert np.isclose(result, expected, rtol=0.05)
        # to get pi in Pa, need V_w in m3/mol
        osmotic_pressure = -8.314 * 298.15 / 0.000018015 * np.log(result)
        assert np.isclose(sol.osmotic_pressure.to("Pa").magnitude, osmotic_pressure, rtol=0.05), f"{osmotic_pressure}"



"""
Tests of the logging module
"""

import logging

from pyEQL.solution import Solution


def test_logging_kwarg(caplog):
    module_log = logging.getLogger("pyEQL")
    with caplog.at_level(logging.DEBUG, "pyEQL"):
        Solution({"Na+": "4 mol/L", "Cl-": "4 mol/L"}, log_level="debug")
        assert module_log.level == logging.DEBUG
        assert "DEBUG" in caplog.text
    caplog.clear()
    with caplog.at_level(logging.ERROR, "pyEQL"):
        Solution({"Na+": "4 mol/L", "Cl-": "4 mol/L"}, log_level="ERROR")
        assert module_log.level == logging.ERROR
        assert "DEBUG" not in caplog.text



"""
pyEQL debye length test suite
============================================

This file contains tests that check the Debye Length
computations of pyEQL

"""

import numpy as np
import pyEQL

# relative tolerance between experimental and computed properties for this test file
RTOL = 0.03


class Test_debye_length:
    """
    test the Debye Length calculations of various solutions
    ------------------------------------------------

    Reference: [1] M. Hu, B. Mi, Enabling graphene oxide nanosheets as water separation membranes,
    Environ. Sci. Technol. 47 (2013) 3715-3723. doi:10.1021/es400571g.

    0.1 mM NaCl: 31nm
    10 mM NaCl: 3.1 nm
    0.1 mM Na2SO4: 18nm
    10 mM Na2SO4: 1.8nm

    """

    def test_debye_length_1(self):
        """ """
        s1 = pyEQL.Solution([["Na+", "0.1 mmol/L"], ["Cl-", "0.1 mmol/L"]])

        result = s1.debye_length.magnitude
        expected = 31

        assert np.isclose(result, expected, rtol=RTOL)

    def test_debye_length_2(self):
        """ """
        s1 = pyEQL.Solution([["Na+", "10 mmol/L"], ["Cl-", "10 mmol/L"]])

        result = s1.debye_length.magnitude
        expected = 3.1

        assert np.isclose(result, expected, rtol=RTOL)

    def test_debye_length_3(self):
        """ """
        s1 = pyEQL.Solution([["Na+", "0.2 mmol/L"], ["SO4-2", "0.1 mmol/L"]])

        result = s1.debye_length.magnitude
        expected = 18

        assert np.isclose(result, expected, rtol=RTOL)

    def test_debye_length_4(self):
        """ """
        s1 = pyEQL.Solution([["Na+", "20 mmol/L"], ["SO4-2", "10 mmol/L"]])

        result = s1.debye_length.magnitude
        expected = 1.8

        assert np.isclose(result, expected, rtol=RTOL)
